import os
import time
import pytest
import signal
import subprocess
import coverage.data
import urllib.request

from threading import Timer
from urllib.error import URLError

from ocs.ocs_client import OCSClient


SIGINT_TIMEOUT = 10


class _AgentRunner:
    """Class to manage running an agent as a subprocess during testing.

    Parameters:
        agent_path (str): Relative path to Agent,
            i.e. '../agents/fake_data/fake_data_agent.py'
        agent_name (str): Short, unique name for the agent
        args (list): Additional CLI arguments to add when starting the Agent

    """

    def __init__(self, agent_path, agent_name, args):
        self.env = os.environ.copy()
        self.env['COVERAGE_FILE'] = f'.coverage.agent.{agent_name}'
        self.env['OCS_CONFIG_DIR'] = os.getcwd()
        self.cmd = [
            'python',
            '-u',
            '-m',
            'coverage',
            'run',
            '--rcfile=./.coveragerc',
            agent_path,
            '--site-file',
            './default.yaml'
        ]
        if args is not None:
            self.cmd.extend(args)
        self.agent_name = agent_name
        self.proc = None
        self._timer = None
        self._timedout = False

    def run(self, timeout):
        """Run the agent subprocess.

        This runs the agent subprocess defined by ``self.cmd``. Output is
        written to a ``PIPE``. If the agent does not exit within the given
        timeout it will be interrupted with a ``SIGKILL``.

        Parameters:
            timeout (float): Timeout in seconds to wait for agent to exit.

        """
        self.proc = subprocess.Popen(self.cmd,
                                     env=self.env,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     text=True,
                                     preexec_fn=os.setsid)

        # start timer for if agent crashes and hangs
        self._timer = Timer(timeout, self._interrupt)
        self._timer.start()

        # Wait briefly then make sure subprocess hasn't already exited.
        time.sleep(1)
        if self.proc.poll() is not None:
            self._timer.cancel()
            self._raise_subprocess(f"Agent failed to startup, cmd: {self.cmd}")

    def _interrupt(self):
        # not graceful, but handles really misbehaved agent subprocesses
        self.proc.send_signal(signal.SIGKILL)
        self._timedout = True

    def _raise_subprocess(self, msg):
        stdout, stderr = self.proc.stdout.read(), self.proc.stderr.read()
        print(f'Here is stdout from {self.agent_name}:\n{stdout}')
        print(f'Here is stderr from {self.agent_name}:\n{stderr}')
        raise RuntimeError(msg)

    def shutdown(self):
        """Shutdown the agent process.

        If the agent does not respond to a ``SIGINT`` then output is printed,
        and an exception raised.

        """
        # don't send SIGINT if we've already sent SIGKILL
        if not self._timedout:
            self.proc.send_signal(signal.SIGINT)
        self._timer.cancel()

        try:
            self.proc.communicate(timeout=SIGINT_TIMEOUT)
        except subprocess.TimeoutExpired:
            self._raise_subprocess('Agent did not terminate within '
                                   f'{SIGINT_TIMEOUT} seconds on SIGINT.')

        if self._timedout:
            stdout, stderr = self.proc.communicate(timeout=SIGINT_TIMEOUT)
            print(f'Here is stdout from {self.agent_name}:\n{stdout}')
            print(f'Here is stderr from {self.agent_name}:\n{stderr}')
            raise RuntimeError('Agent timed out.')


def create_agent_runner_fixture(agent_path, agent_name, args=None, timeout=60):
    """Create a pytest fixture for running a given OCS Agent.

    Parameters:
        agent_path (str): Relative path to Agent,
            i.e. '../agents/fake_data/fake_data_agent.py'
        agent_name (str): Short, unique name for the agent
        args (list): Additional CLI arguments to add when starting the Agent
        timeout (float): Timeout in seconds, after which the agent process will
            be interrupted. This typically indicates a crash within the agent.
            This timeout should be longer than you expect the agent to run for
            during a given test. Defaults to 60 seconds.

    """
    @pytest.fixture()
    def run_agent(cov):
        runner = _AgentRunner(agent_path, agent_name, args)
        runner.run(timeout=timeout)

        yield

        runner.shutdown()

        # report coverage
        agentcov = coverage.data.CoverageData(
            basename=f'.coverage.agent.{agent_name}')
        agentcov.read()
        # protect against missing --cov flag
        if cov is not None:
            cov.get_data().update(agentcov)

    return run_agent


def create_client_fixture(instance_id, timeout=30):
    """Create the fixture that provides tests a Client object.

    Parameters:
        instance_id (str): Agent instance-id to connect the Client to
        timeout (int): Approximate timeout in seconds for the connection.
            Connection attempts will be made X times, with a 1 second pause
            between attempts. This is useful if it takes some time for the
            Agent to start accepting connections, which varies depending on the
            Agent.

    """
    @pytest.fixture()
    def client_fixture():
        # Set the OCS_CONFIG_DIR so we read the local default.yaml file
        os.environ['OCS_CONFIG_DIR'] = os.getcwd()
        print(os.environ['OCS_CONFIG_DIR'])
        attempts = 0

        while attempts < timeout:
            try:
                client = OCSClient(instance_id)
                return client
            except RuntimeError as e:
                print(f"Caught error: {e}")
                print("Attempting to reconnect.")

            time.sleep(1)
            attempts += 1

        raise RuntimeError(
            f"Failed to connect to {instance_id} after {timeout} attempts.")

    return client_fixture


def check_crossbar_connection(port=18001, interval=5, max_attempts=6):
    """Check that the crossbar server is up and available for an Agent to
    connect to.

    Parameters:
        port (int): Port the crossbar server is configured to run on for
            testing.
        interval (float): Amount of time in seconds to wait between checks.
        max_attempts (int): Maximum number of attempts before giving up.

    Notes:
        For this check to work the crossbar server needs the `Node Info Service
        <https://crossbar.io/docs/Node-Info-Service/>`_ running at the path
        /info.

    """
    attempts = 0

    while attempts < max_attempts:
        try:
            url = f"http://localhost:{port}/info"
            code = urllib.request.urlopen(url).getcode()
        except (URLError, ConnectionResetError):
            print("Crossbar server not online yet, waiting 5 seconds.")
            time.sleep(interval)

        attempts += 1

    assert code == 200
    print("Crossbar server online.")



# The crossbar router must be configured to expose http access.
import json
import requests


class ControlClientError(RuntimeError):
    pass


class ControlClient():
    def __init__(self, agent_addr, **kwargs):
        self.agent_addr = agent_addr
        self.realm = kwargs['realm']
        self.call_url = kwargs['url']

    # start and stop are just to imitate wampy client...
    def start(self, *args, **kwargs):
        pass

    def stop(self, *args, **kwargs):
        pass

    def call(self, procedure, *args, **kwargs):
        # curl -H "Content-Type: application/json"
        #     -d '{"procedure": "observatory.acu1",
        #          "args": ["get_tasks"]}'
        #     http://127.0.0.1:8001/call
        params = json.dumps({'procedure': procedure,
                             'args': args, 'kwargs': kwargs})
        try:
            r = requests.post(self.call_url, data=params)
        except requests.exceptions.ConnectionError:
            raise ControlClientError([0, 0, 0, 0, 'client_http.error.connection_error',
                                      ['Failed to connect to %s' % self.call_url], {}])
        if r.status_code != 200:
            raise ControlClientError([0, 0, 0, 0, 'client_http.error.request_error',
                                      ['Server replied with code %i' % r.status_code], {}])
        decoded = r.json()
        if 'error' in decoded:
            # Return errors in the same way wampy does, roughly.
            raise ControlClientError([0, 0, 0, 0, decoded['error'], decoded['args'], decoded['kwargs']])
        return decoded['args'][0]

    def get_api(self, simple=False):
        """Query the API and other info from the Agent; this includes lists of
        Processes, Tasks, and Feeds, docstrings, operation session
        structures, and info about the Agent instance (class, PID,
        host).

        Args:
          simple (bool): If True, then return just the lists of the op
            and feed names without accompanying detail.

        Returns:
          A dict, see :func:`ocs.ocs_agent.OCSAgent._management_handler`
          for detail.

        """
        data = self.call(self.agent_addr, 'get_api')
        if not simple:
            return data
        return {k: [_v[0] for _v in v]
                for k, v in data.items() if isinstance(v, dict)}

    def get_tasks(self):
        """
        Query the list of Tasks from the Agent management interface.

        Returns a list of items of the form (task_name, info_dict).
        """
        return self.call(self.agent_addr, 'get_tasks')

    def get_processes(self):
        """
        Query the list of Processes from the Agent management interface.

        Returns a list of items of the form (process_name, info_dict).
        """
        return self.call(self.agent_addr, 'get_processes')

    def get_feeds(self):
        """
        Query the list of Feeds from the Agent management interface.

        Returns a list of items of the form (feed_name, info_dict).
        """
        return self.call(self.agent_addr, 'get_feeds')

    def request(self, action, op_name, params={}, **kw):
        """
        Issue a request on an Agent's .ops interface.

        Args:
          action (string): The action name (start, status, etc).
          params (dict): Parameters to pass to the action.

        Returns:
          Tuple (status, message, session).
        """
        return self.call(self.agent_addr + '.ops', action, op_name, params, **kw)



from ocs.ocs_agent import in_reactor_context
from twisted.internet import reactor
from autobahn.wamp.exception import TransportLost
import time
import re


class Block:
    def __init__(self, name, keys):
        """
        Structure of block for a so3g IrregBlockDouble.
        """
        self.name = name
        self.timestamps = []
        self.data = {
            k: [] for k in keys
        }

    def empty(self):
        """ Returns true if block is empty"""
        return self.timestamps == []

    def clear(self):
        """
        Empties block's buffers
        """
        self.timestamps = []
        for key in self.data:
            self.data[key] = []

    def append(self, d):
        """
        Adds a single data point to the block
        """
        if d['data'].keys() != self.data.keys():
            raise Exception("Block structure does not match: {}".format(self.name))

        self.timestamps.append(d['timestamp'])

        for k in self.data:
            self.data[k].append(d['data'][k])

    def extend(self, block):
        """
        Extends the data block by an encoded block
        """
        if block['data'].keys() != self.data.keys():
            raise Exception("Block structure does not match: {}".format(self.name))

        self.timestamps.extend(block['timestamps'])
        for k in self.data:
            self.data[k].extend(block['data'][k])

    def encoded(self):
        n = len(self.timestamps)
        assert (all([n == len(v) for v in self.data.values()]))
        return {
            'block_name': self.name,
            'data': {k: self.data[k] for k in self.data.keys()},
            'timestamps': self.timestamps,
        }


class Feed:
    """
    Manages publishing to a specific feed and storing of messages.

    Args:
        agent (OCSAgent):
            agent that is registering the feed
        feed_name (string):
            name of the feed
        record (bool, optional):
            Determines if feed should be aggregated. At the moment, each agent
            can have at most one aggregated feed. Defaults to False
        agg_params (dict, optional):
            Parameters used by the aggregator and influx publisher.

            Params:
                **frame_length** (float):
                    Deterimes the amount of time each G3Frame should be (in seconds).
                **fresh_time** (float):
                    Time until feed is considered stale by aggregator.
                **exclude_aggregator** (bool):
                    If True, the HK Aggregator will not record feed to g3.
                **exclude_influx** (bool):
                    If True, the InfluxPublisher will not write the feed to
                    Influx.

        buffer_time (int, optional):
            Specifies time that messages should be buffered in seconds.
            If 0, message will be published immediately.
            Defaults to 0.
        max_messages (int, optional):
            Max number of messages stored. Defaults to 20.
    """

    def __init__(self, agent, feed_name, record=False, agg_params={},
                 buffer_time=0, max_messages=0):

        self.agent = agent
        self.feed_name = feed_name
        self.record = record
        self.agg_params = agg_params

        self.buffer_time = buffer_time
        self.buffer_start_time = None

        self.blocks = {}

        self.agent_address = self.agent.agent_address
        self.address = "{}.feeds.{}".format(self.agent_address, self.feed_name)

    def encoded(self):
        return {
            "agent_address": self.agent_address,
            "agg_params": self.agg_params,
            "feed_name": self.feed_name,
            "address": self.address,
            "record": self.record,
            "session_id": self.agent.agent_session_id,
            "agent_class": self.agent.class_name
        }

    def flush_buffer(self):
        """Publishes all messages in buffer and empties it."""

        if not in_reactor_context():
            return reactor.callFromThread(self.flush_buffer)
        if self.buffer_start_time is None:
            return

        if self.record:
            try:
                self.agent.publish(self.address, (
                    {k: b.encoded() for k, b in self.blocks.items() if b.timestamps},
                    self.encoded()
                ))
            except TransportLost:
                self.agent.log.error('Could not publish to Feed. TransportLost. '
                                     + 'crossbar server likely unreachable.')
            for k, b in self.blocks.items():
                b.clear()

    def publish_message(self, message, timestamp=None):
        """
        Publishes message to feed.  If this is an aggregatable feed
        (record=True), then it may be buffered.  Otherwise it is
        dispatched immediately.

        Args:
            message:
                Data to be published (see notes about acceptable formats).
            timestamp (float):
                timestamp given to the message. Defaults to time.time()

        If this feed is not intended to provide structured data for
        aggregation, then the format of the message is unrestricted as
        long as it is WAMP-serializable.

        For aggregated feeds, the message should be a dict with one of
        the following formats:

        1. A single sample for several co-sampled channels.  The
           structure is::

             message = {
                 'block_name': Key given to the block in blocking param
                 'timestamp': timestamp of data
                 'data': {
                      key1: datapoint1
                      key2: datapoint2
                  }
             }

           Samples recorded in this way may be buffered, if
           self.sample_time > 0.

        2. Multiple or more samples for several co-sampled channels.
           The structure is::


             message = {
                 'block_name': Key given to the block in blocking param
                 'timestamps': [timestamp, timestamp...]
                 'data': {
                      key1: [datapoint, datapoint...]
                      key2: [datapoint, datapoint...]
                  }
             }

           Note that the code distinguishes between these cases based
           on the presence of the key 'timestamps' rather than
           'timestamp'.  These data can be buffered, too, if
           self.sample_time > 0.

        """
        current_time = time.time()

        if timestamp is None:
            timestamp = current_time

        if not in_reactor_context():
            # Take a copy, for thread-safety.
            message = message.copy()
            return reactor.callFromThread(self.publish_message, message,
                                          timestamp=timestamp)

        if self.record:
            # check message contents
            for k, v in message['data'].items():
                Feed.verify_data_field_string(k)
                Feed.verify_message_data_type(v)

            # Data is stored in Block objects
            block_name = message['block_name']
            try:
                b = self.blocks[block_name]
            except KeyError:
                b = Block(block_name, message['data'].keys())
                self.blocks[block_name] = b

            if 'timestamp' in message:
                b.append(message)
            elif 'timestamps' in message:
                b.extend(message)
            else:
                raise RuntimeError('Invalid message when record=True.  keys=%s' %
                                   message.keys())

            if self.buffer_start_time is None:
                self.buffer_start_time = current_time

            if (current_time - self.buffer_start_time) >= self.buffer_time:
                self.flush_buffer()
                self.buffer_start_time = None

        else:
            # Publish message immediately
            try:
                self.agent.publish(self.address, (message, self.encoded()))
            except TransportLost:
                self.agent.log.error('Could not publish to Feed. TransportLost. '
                                     + 'crossbar server likely unreachable.')

    @staticmethod
    def verify_message_data_type(value):
        """Aggregated Feeds can only store certain types of data. Here we check
        that the type of all data contained in a message's 'data' dictionary are
        supported types.

        Args:
            value (list, float, int, bool):
                'data' dictionary value published (see Feed.publish_message for details).

        """
        valid_types = (float, int, str, bool)

        # multi-sample check
        if isinstance(value, list):
            if not all(isinstance(x, valid_types) for x in value):
                type_set = set([type(x) for x in value])
                invalid_types = type_set.difference(valid_types)
                raise TypeError("message 'data' block contains invalid data"
                                + f"types: {invalid_types}")

        # single sample check
        else:
            if not isinstance(value, valid_types):
                invalid_type = type(value)
                raise TypeError("message 'data' block contains invalid "
                                + f"data type: {invalid_type}")

    @staticmethod
    def verify_data_field_string(field):
        """There are strict rules for the characters allowed in field names.
        This function verifies the names in a message are valid.

        A valid name:

        * contains only letters (a-z, A-Z; case sensitive), decimal digits (0-9), and the
          underscore (_).
        * begins with a letter, or with any number of underscores followed by a letter.
        * is at least one, but no more than 255, character(s) long.

        Args:
            field (str):
                Field name string to verify.

        Returns:
            bool: True if field name is valid.

        Raises:
            ValueError: If field name is invalid.

        """
        # Check for empty field name
        if not field:
            raise ValueError("Empty field name encountered, please enter "
                             + "a valid field name.")

        # Complement (^) the set, matching any unlisted characters
        check_invalid = re.compile('[^a-zA-Z0-9_]')

        # Similar to check_invalid, search for non letter characters
        # Leading ^ matches the start of string, so following numbers are valid
        check_start = re.compile('^[^a-zA-Z]')

        # check for invalid characters
        result = check_invalid.search(field)
        if result:
            raise ValueError(f"message 'data' block contains the key {field} "
                             f"with the invalid character '{result.group(0)}'. "
                             "Valid characters are a-z, A-Z, 0-9, and underscore.")

        # check for non-letter start, even after underscores
        stripped_key = field.strip("_")
        if check_start.search(stripped_key):
            raise ValueError(f"message 'data' block contains the key {field}, "
                             + "which does not start with a letter (after any "
                             + "number of leading underscores.)")

        # check key length
        if len(field) > 255:
            raise ValueError(f"message 'data' block contains key {field} which "
                             + "exceeds the valid length of 255 characters.")

        return True

    @staticmethod
    def enforce_field_name_rules(field_name):
        """Enforce naming rules for field names.

        A valid name:

        * contains only letters (a-z, A-Z; case sensitive), decimal digits (0-9), and the
          underscore (_).
        * begins with a letter, or with any number of underscores followed by a letter.
        * is at least one, but no more than 255, character(s) long.

        Args:
            field_name (str):
                Field name string to check and modify if needed.

        Returns:
            str: New field name, meeting all above rules. Note this isn't
                 guarenteed to not collide with other field names passed
                 through this method, and that should be checked.

        """
        # check for empty string
        if field_name == "":
            new_field_name = "invalid_field"
        else:
            new_field_name = field_name

        # replace spaces with underscores
        new_field_name = new_field_name.replace(' ', '_')

        # replace invalid characters
        new_field_name = re.sub('[^a-zA-Z0-9_]', '', new_field_name)

        # grab leading underscores
        underscore_search = re.compile('^_*')
        underscores = underscore_search.search(new_field_name).group()

        # remove leading underscores
        new_field_name = re.sub('^_*', '', new_field_name)

        # remove leading non-letters
        new_field_name = re.sub('^[^a-zA-Z]*', '', new_field_name)

        # add underscores back
        new_field_name = underscores + new_field_name

        # limit to 255 characters
        new_field_name = new_field_name[:255]

        return new_field_name



import time
import queue
import argparse
import txaio

from os import environ

from ocs import ocs_agent, site_config
from ocs.base import OpCode

from ocs.agents.influxdb_publisher.drivers import Publisher

# For logging
txaio.use_twisted()
LOG = txaio.make_logger()


class InfluxDBAgent:
    """
    This class provide a WAMP wrapper for the data publisher. The run function
    and the data handler **are** thread-safe, as long as multiple run functions
    are not started at the same time, which should be prevented through
    OCSAgent.

    Args:
        agent (OCSAgent):
            OCS Agent object
        args (namespace):
            args from the function's argparser.

    Attributes:
        data_dir (path):
            Path to the base directory where data should be written.
        aggregate (bool):
           Specifies if the agent is currently aggregating data.
        incoming_data (queue.Queue):
            Thread-safe queue where incoming (data, feed) pairs are stored
            before being passed to the Publisher.
        loop_time (float):
            Time between iterations of the run loop.
    """

    def __init__(self, agent, args):
        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.args = args

        self.aggregate = False
        self.incoming_data = queue.Queue()
        self.loop_time = 1

        self.agent.subscribe_on_start(self._enqueue_incoming_data,
                                      f'{args.address_root}..feeds.',
                                      options={'match': 'wildcard'})

        record_on_start = (args.initial_state == 'record')
        self.agent.register_process('record',
                                    self.record, self._stop_record,
                                    startup=record_on_start)

    def _enqueue_incoming_data(self, _data):
        """Data handler for all feeds. This checks to see if the feeds should
        be recorded, and if they are it puts them into the incoming_data queue
        to be processed by the Publisher during the next run iteration.

        """
        data, feed = _data

        if not feed['record'] or not self.aggregate:
            return

        # LOG.debug("data: {d}", d=data)
        # LOG.debug("feed: {f}", f=feed)

        self.incoming_data.put((data, feed))

    @ocs_agent.param('test_mode', default=False, type=bool)
    def record(self, session: ocs_agent.OpSession, params):
        """record()

        **Process** - This process will create an Publisher instance, which
        will collect and write provider data to disk as long as this process is
        running.

        Parameters:
            test_mode (bool, optional): Run the record Process loop only once.
                This is meant only for testing. Default is False.

        """
        self.aggregate = True

        self.log.debug("Instatiating Publisher class")
        publisher = Publisher(self.args.host,
                              self.args.database,
                              self.incoming_data,
                              port=self.args.port,
                              protocol=self.args.protocol,
                              gzip=self.args.gzip,
                              operate_callback=lambda: self.aggregate,
                              )

        while self.aggregate:
            time.sleep(self.loop_time)
            self.log.debug(f"Approx. queue size: {self.incoming_data.qsize()}")
            publisher.run()

            if params['test_mode']:
                break

        publisher.close()

        return True, "Aggregation has ended"

    def _stop_record(self, session, params):
        if OpCode(session.op_code) in [OpCode.STARTING, OpCode.RUNNING]:
            session.set_status('stopping')
            self.aggregate = False
            return True, "Stopping aggregation"
        elif OpCode(session.op_code) == OpCode.STOPPING:
            return True, "record process status is already 'stopping'"
        else:
            return False, "record process not currently running"


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--initial-state',
                        default='record', choices=['idle', 'record'],
                        help="Initial state of argument parser. Can be either "
                             "idle or record")
    pgroup.add_argument('--host',
                        default='influxdb',
                        help="InfluxDB host address.")
    pgroup.add_argument('--port',
                        default=8086,
                        help="InfluxDB port.")
    pgroup.add_argument('--database',
                        default='ocs_feeds',
                        help="Database within InfluxDB to publish data to.")
    pgroup.add_argument('--protocol',
                        default='line',
                        choices=['json', 'line'],
                        help="Protocol for writing data. Either 'line' or "
                             "'json'.")
    pgroup.add_argument('--gzip',
                        type=bool,
                        default=False,
                        help="Use gzip content encoding to compress requests.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='InfluxDBAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    InfluxDBAgent(agent, args)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import ocs
from ocs import ocs_agent, site_config, agent_cli
from ocs.agents.host_manager import drivers as hm_utils

import time
import argparse

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue
from autobahn.twisted.util import sleep as dsleep

import os
import sys

VALID_TARGETS = ['up', 'down']

# "agent_class" value used for docker services that do not seem to
# have a corresponding SCF entry.
NONAGENT_DOCKER = '[docker]'


class HostManager:
    """
    This Agent is used to start and stop OCS-relevant services on a
    particular host.  If the HostManager is launched automatically when
    a system boots, it can then be used to start up the rest of OCS on
    that host (either automatically or on request).
    """

    def __init__(self, agent, docker_composes=[],
                 docker_service_prefix='ocs-'):
        self.agent = agent
        self.running = False
        self.database = {}  # key is instance_id (or docker service name).
        self.docker_composes = docker_composes
        self.docker_service_prefix = docker_service_prefix

    @inlineCallbacks
    def _get_local_instances(self):
        """Parse the site config and return a list of this host's Agent
        instances.

        Returns:
          success (bool): True if config was successfully scanned.
            False otherwise, indiating perhaps we should go into a bit
            of a lock down while operator sorts that out.
          agent_dict (dict): Maps instance-id to a dict describing the
            agent config.  The config is as contained in
            HostConfig.instances but where 'instance-id',
            'agent-class', and 'manage' are all guaranteed populated
            (and manage is a valid full description, e.g. "host/down").
          warnings: A list of strings, each of which corresponds to
            some problem found in the config.

        """
        warnings = []
        instances = {}

        # Load site config file.
        try:
            site, hc, _ = site_config.get_config(
                self.agent.site_args, '*host*')
            self.site_config_file = site.source_file
            self.host_name = hc.name
            self.working_dir = hc.working_dir
        except Exception as e:
            warnings.append('Failed to read site config file -- '
                            f'likely syntax error: {e}')
            return returnValue((False, instances, warnings))

        # Scan for agent scripts in (deprecated) script registry
        try:
            for p in hc.agent_paths:
                if p not in sys.path:
                    sys.path.append(p)
            site_config.scan_for_agents()
        except Exception as e:
            warnings.append('Failed to scan for old plugin agents -- '
                            f'likely plugin config problem: {e}')
            return returnValue((False, instances, warnings))

        # Gather managed items from site config.
        for inst in hc.instances:
            if inst['instance-id'] in instances:
                warnings.append(
                    f'Configuration problem, instance-id={inst["instance-id"]} '
                    f'has multiple entries.  Ignoring repeats.')
                continue
            if inst['agent-class'] == 'HostManager':
                inst['manage'] = 'ignore'
            else:
                # Make sure 'manage' is set, and valid.
                inst['manage'] = inst.get('manage', None)
                try:
                    inst['manage'] = site_config.InstanceConfig._MANAGE_MAP[inst['manage']]
                except KeyError:
                    warnings.append(
                        f'Configuration problem, invalid manage={inst["manage"]} '
                        f'for instance_id={inst["instance-id"]}.')
                    continue
            instances[inst['instance-id']] = inst
        returnValue((True, instances, warnings))
        yield

    @inlineCallbacks
    def _update_docker_services(self, session):
        """Parse the docker-compose.yaml files and return the current
        status of all services.  For any services matching
        self.database entries, the corresponding DockerContainerHelper
        is updated with the new info.

        Returns:
          docker_services (dict): state information for all detected
            services, keyed by the service name.

        """
        # Read services from all docker-compose files.
        docker_services = {}
        for compose in self.docker_composes:
            try:
                services = yield hm_utils.parse_docker_state(compose)
                this_ok = True
                this_msg = f'Successfully parsed {compose} and its service states.'
            except Exception as e:
                this_ok = False
                this_msg = (f'Failed to interpret {compose} and/or '
                            f'its service states: {e}')

            # Don't issue the same complaint more than once per minute or so
            compose_was_ok, timestamp, last_msg = self.config_parse_status.get(
                compose, (False, 0, ''))
            if (this_ok != compose_was_ok) \
               or (not this_ok and time.time() - timestamp > 60) \
               or (not this_ok and this_msg != last_msg):
                session.add_message(this_msg)
                self.config_parse_status[compose] = (this_ok, time.time(), this_msg)

            if this_ok:
                docker_services.update(services)

        # Update all docker things in the database.
        retirees = []
        assigned_services = []
        for key, instance in self.database.items():
            if instance['management'] != 'docker':
                continue

            service_name = instance['agent_script']
            service_data = docker_services.get(service_name)
            assigned_services.append(service_name)

            prot = instance.get('prot')
            if prot is None:
                if service_data is not None:
                    # Create a prot entry with the service info.
                    instance['prot'] = hm_utils.DockerContainerHelper(service_data)
                    instance['operable'] = True
                    if instance['agent_class'] != NONAGENT_DOCKER:
                        instance['agent_class'] = _clsname_tool(instance['agent_class'], '[d]')
            else:
                if service_data is not None:
                    prot.update(service_data)
                else:
                    # service_data is missing, but there used to be a
                    # service there.  Close it out.
                    instance['prot'] = None
                    instance['operable'] = False
                    if instance['agent_class'] == NONAGENT_DOCKER:
                        session.add_message(f'Deleting non-agent service {key}')
                        retirees.append(key)
                    else:
                        session.add_message(f'Marking missing service for {key}')
                        instance['agent_class'] = _clsname_tool(instance['agent_class'], '[d?]')

        # If a non-agent [docker] service has disappeared, there's no
        # reason to show it in a list, and no persistent state /
        # operations to worry about.  So just delete it.
        for r in retirees:
            self.database.pop(r)

        # Create entries for any new un-matched docker services.
        unassigned_services = set(docker_services.keys()) \
            .difference(assigned_services)
        for srv in unassigned_services:
            instance = hm_utils.ManagedInstance.init(
                management='docker',
                instance_id=srv,
                agent_class=NONAGENT_DOCKER,
                full_name=(f'[docker]:{srv}'))
            instance.update({
                'agent_script': srv,
                'operable': True,
            })
            service_data = docker_services[srv]
            instance['prot'] = hm_utils.DockerContainerHelper(service_data)

            self.database[srv] = instance
            # If it's up, leave it up.
            if service_data['running']:
                instance['target_state'] = 'up'
                instance['next_action'] = 'up'

        returnValue(docker_services)

    @inlineCallbacks
    def _reload_config(self, session):
        """This helper function is called by both the ``manager``
        Process at startup, and the ``update`` Task.

        The Site Config File is parsed and used to update the internal
        database of child instances.  Any previously unknown child
        Agent is added to the internal tracking database, and assigned
        whatever target state is specified for that instance.  Any
        previously known child Agent instance is not modified.

        If any child Agent instances in the internal database appear
        to have been removed from the SCF, then they are set to have
        target_state "down" and will be deleted from the database when
        that state is reached.

        """
        def retire(db_key):
            instance = self.database.get(db_key, None)
            if instance is None:
                return
            instance['management'] = 'retired'
            instance['at'] = time.time()
            instance['target_state'] = 'down'

        def _full_name(cls, iid):
            if cls != NONAGENT_DOCKER:
                cls, _ = _clsname_tool(cls)
            return f'{cls}:{iid}'

        def same_base_class(a, b):
            return _clsname_tool(a)[0] == _clsname_tool(b)[0]

        # Parse the site config.
        parse_ok, agent_dict, warnings = yield self._get_local_instances()
        for w in warnings:
            session.add_message(w)

        self.config_parse_status['[SCF]'] = (parse_ok, time.time(), ''.join(warnings))
        if not parse_ok:
            return warnings

        # Any agents in the database that are not listed in the latest
        # agent_dict should be immediately retired.  That includes
        # things that are suddenly marked as manage=no.  Ignore docker
        # non-agents.
        for iid, instance in self.database.items():
            if instance['agent_class'] != NONAGENT_DOCKER and (
                    iid not in agent_dict or agent_dict[iid].get('manage') == 'ignore'):
                session.add_message(
                    f'Retiring {instance["full_name"]}, which has disappeared from '
                    f'configuration file(s) or has manage:no.')
                retire(iid)

        # Create / update entries for every agent in the host
        # description, unless it is explicitly marked as ignore.
        for iid, hinst in agent_dict.items():
            if hinst['manage'] == 'ignore':
                continue

            cls = hinst['agent-class']
            srv = None  # The expected docker service name, if any

            mgmt, start_state = hinst['manage'].split('/')
            if mgmt == 'docker':
                cls = _clsname_tool(cls, '[d?]')
                srv = self.docker_service_prefix + iid

            # See if we already tracking this agent.
            instance = self.database.get(iid)

            if instance is not None:
                # Already tracking; just check for major config change.
                _cls = instance['agent_class']
                _mgmt = instance['management']
                if not same_base_class(_cls, cls) or _mgmt != mgmt:
                    session.add_message(
                        f'Managed agent "{iid}" changed agent_class '
                        f'({_cls} -> {cls}) or management '
                        f'({_mgmt} -> {mgmt}) and is being reset!')
                    # Bring down existing instance
                    self._terminate_instance(iid)
                    # Start a new one
                    instance = None

            # Do we have an unmatched docker entry for this?
            if instance is None and srv in self.database:
                # Re-register it under instance_id
                instance = self.database.pop(srv)
                self.database[iid] = instance
                instance.update({
                    'instance_id': iid,
                    'agent_class': _clsname_tool(cls, '[d]'),
                    'full_name': _full_name(cls, iid),
                })

            if instance is None:
                instance = hm_utils.ManagedInstance.init(
                    management=mgmt,
                    instance_id=iid,
                    agent_class=cls,
                    full_name=_full_name(cls, iid),
                )
                instance['target_state'] = start_state
                self.database[iid] = instance

        # Get agent class list from modern plugin system.
        agent_plugins = agent_cli.build_agent_list()

        # Assign plugins / scripts / whatever to any new instances.
        for iid, instance in self.database.items():
            if instance['agent_script'] is not None:
                continue
            if instance['management'] == 'host':
                cls = instance['agent_class']
                # Check for the agent class in the plugin system;
                # then check the (deprecated) agent script registry.
                if cls in agent_plugins:
                    session.add_message(f'Found plugin for "{cls}"')
                    instance['agent_script'] = '__plugin__'
                    instance['operable'] = True
                elif cls in site_config.agent_script_reg:
                    session.add_message(f'Found launcher script for "{cls}"')
                    instance['agent_script'] = site_config.agent_script_reg[cls]
                    instance['operable'] = True
                else:
                    session.add_message(f'No plugin (nor launcher script) '
                                        f'found for agent_class "{cls}"!')
            elif instance['management'] == 'docker':
                instance['agent_script'] = self.docker_service_prefix + iid

        # Read the compose files; query container states; updater stuff.
        yield self._update_docker_services(session)
        returnValue(warnings)

    def _launch_instance(self, instance):
        """Launch an Agent instance (whether 'host' or 'docker' managed) using
        hm_utils.

        For 'host' managed agents: hm_utils will use
        reactor.spawnProcess, and store the AgentProcessHelper (which
        inherits from twisted ProcessProtocol) in instance['prot'].
        The site_file and instance_id are passed on the command line;
        this means that any weird config overrides passed to this
        HostManager are not propagated.  One exception is working_dir,
        which is propagated in order that relative paths can make any
        sense.

        For 'docker' managed agents: hm_utils will try to start the
        right service container, and instance['prot'] will hold a
        DockerContainerHelper (which has some common interface with
        AgentProcessHelper).

        """
        if instance['management'] == 'docker':
            prot = instance['prot']
        else:
            iid = instance['instance_id']
            pyth = sys.executable
            script = instance['agent_script']
            if script == '__plugin__':
                cmd = [pyth, '-m', 'ocs.agent_cli']
            else:
                cmd = [pyth, script]
            cmd.extend([
                '--instance-id', iid,
                '--site-file', self.site_config_file,
                '--site-host', self.host_name,
                '--working-dir', self.working_dir])
            prot = hm_utils.AgentProcessHelper(iid, cmd)
        prot.up()
        instance['prot'] = prot

    def _terminate_instance(self, key):
        """
        Use the ProcessProtocol to request the Agent instance to exit.
        """
        prot = self.database[key]['prot']  # Get the ProcessProtocol.
        if prot is None:
            return True, 'Instance was not running.'
        if prot.killed:
            return True, 'Instance already has kill set.'
        # Note the .down call does not block -- the thing will be
        # stopped from a thread.
        prot.down()
        return True, 'Kill requested.'

    def _process_target_states(self, session, requests=[]):
        """This is a helper function for parsing target_state change
        requests; see the update Task.

        """
        # Special requests will target specific instance_id; make a map for that.
        addressable = {k: v for k, v in self.database.items()
                       if v['management'] != 'retired'}

        for key, state in requests:
            if state not in VALID_TARGETS:
                session.add_message('Ignoring request for "%s" -> invalid state "%s".' %
                                    (key, state))
                continue
            if key == 'all':
                for v in addressable.values():
                    v['target_state'] = state
            else:
                if key in addressable:
                    addressable[key]['target_state'] = state
                else:
                    session.add_message(f'Ignoring invalid target, {key}')

    @ocs_agent.param('requests', default=[])
    @ocs_agent.param('reload_config', default=True, type=bool)
    @inlineCallbacks
    def manager(self, session, params):
        """manager(requests=[], reload_config=True)

        **Process** - The "manager" Process maintains a list of child
        Agents for which it is responsible.  In response to requests
        from a client, the Process will launch or terminate child
        Agents.

        Args:

          requests (list): List of agent instance target state
            requests; e.g. [('instance1', 'down')].  See description
            in :meth:`update` Task.
          reload_config (bool): When starting up, discard any cached
            database of tracked agents and rescan the Site Config
            File.  This is mostly for debugging.

        Notes:

          If an Agent process exits unexpectedly, it will be
          relaunched within a few seconds.

          When this Process is started (or restarted), the list of
          tracked agents and their status is completely reset, and the
          Site Config File is read in.

          Once this process is running, the target states for managed
          Agents can be manipulated through the :meth:`update` task.

          Note that when a stop is requested on this Process, all
          managed Agents will be moved to the "down" state and an
          attempt will be made to terminate them before the Process
          exits.

          The session.data is a dict, and entry 'child_states'
          contains a list with the managed Agent statuses.  For
          example::

            {'child_states': [
              {'next_action': 'up',
               'target_state': 'up',
               'stability': 1.0,
               'agent_class': 'Lakeshore372Agent',
               'instance_id': 'thermo1'},
              {'next_action': 'down',
               'target_state': 'down',
               'stability': 1.0,
               'agent_class': 'ACUAgent',
               'instance_id': 'acu-1'},
              {'next_action': 'up',
               'target_state': 'up',
               'stability': 1.0,
               'agent_class': 'FakeDataAgent[d]',
               'instance_id': 'faker6'},
              ],
            }

          If you are looking for the "current state", it's called
          "next_action" here.

          The agent_class may include a suffix [d] or [d?], indicating
          that the agent is configured to run within a docker
          container.  (The question mark indicates that the
          HostManager cannot actually identify the docker-compose
          service associated with the agent description in the SCF.)

        """
        self.config_parse_status = {}
        session.data = {
            'child_states': [],
            'config_parse_status': self.config_parse_status,
        }

        self.running = True

        if params['reload_config']:
            self.database = {}
            yield self._reload_config(session)
        self._process_target_states(session, params['requests'])

        next_docker_update = time.time()

        any_jobs = False
        while self.running or any_jobs:

            if time.time() >= next_docker_update:
                yield self._update_docker_services(session)
                next_docker_update = time.time() + 2

            sleep_times = [1.]
            any_jobs = False

            for key, db in self.database.items():

                # If Process exit is requested, force all targets to down.
                if not self.running:
                    db['target_state'] = 'down'

                actions = hm_utils.resolve_child_state(db)

                for msg in actions['messages']:
                    session.add_message(msg)
                if actions['terminate']:
                    self._terminate_instance(key)
                if actions['launch']:
                    reactor.callFromThread(self._launch_instance, db)
                if actions['sleep']:
                    sleep_times.append(actions['sleep'])
                any_jobs = (any_jobs or (db['next_action'] != 'down'))

                # Criteria for stability:
                db['fail_times'], db['stability'] = hm_utils.stability_factor(
                    db['fail_times'])

            # Clean up retired items.
            self.database = {
                k: v for k, v in self.database.items()
                if v['management'] != 'retired' or v['next_action'] not in ['down', '?']}

            # Update session info.
            child_states = []
            for state in self.database.values():
                child_states.append({_k: state[_k] for _k in
                                     ['next_action',
                                      'target_state',
                                      'stability',
                                      'agent_class',
                                      'instance_id']})
            session.data['child_states'] = child_states

            yield dsleep(max(min(sleep_times), .001))
        return True, 'Exited.'

    @inlineCallbacks
    def _stop_manager(self, session, params):
        yield
        if session.status == 'done':
            return
        session.set_status('stopping')
        self.running = False
        return True, 'Stop initiated.'

    @ocs_agent.param('requests', default=[])
    @ocs_agent.param('reload_config', default=False, type=bool)
    @inlineCallbacks
    def update(self, session, params):
        """update(requests=[], reload_config=False)

        **Task** - Update the target state for any subset of the
        managed agent instances.  Optionally, trigger a full reload of
        the Site Config File first.

        Args:
          requests (list): Default is [].  Each entry must be a tuple
            of the form ``(instance_id, target_state)``.  The
            ``instance_id`` must be a string that matches an item in
            the current list of tracked agent instances, or be the
            string 'all', which will match all items being tracked.
            The ``target_state`` must be 'up' or 'down'.
          reload_config (bool): Default is False.  If True, the site
            config file and docker-compose files are reparsed in order
            to (re-)populate the database of child Agent instances.

        Examples:
          ::

            update(requests=[('thermo1', 'down')])
            update(requests=[('all', 'up')])
            update(reload_config=True)


        Notes:
          Starting and stopping agent instances is handled by the
          :meth:`manager` Process; if that Process is not running then
          no action is taken by this Task and it will exit with an
          error.

          The entries in the ``requests`` list are processed in order.
          For example, if the requests were [('all', 'up'), ('data1',
          'down')].  This would result in setting all known children
          to have target_state "up", except for "data1" which would be
          given target state of "down".

          If ``reload_config`` is True, the Site Config File will be
          reloaded (as described in :meth:`_reload_config`) before
          any of the requests are processed.

          Managed docker-compose.yaml files are reparsed, continously,
          by the manager process -- no specific action is taken with
          those in this Task.  Note that adding/changing the list of
          docker-compose.yaml files requires restarting the agent.

        """
        if not self.running:
            return False, 'Manager process is not running; params not updated.'
        if params['reload_config']:
            yield self._reload_config(session)
        self._process_target_states(session, params['requests'])
        return True, 'Update requested.'

    @inlineCallbacks
    def die(self, session, params):
        if not self.running:
            session.add_message('Manager process is not running.')
        else:
            session.add_message('Requesting exit of manager process.')
            ok, msg, mp_session = self.agent.stop('manager')
            ok, msg, mp_session = yield self.agent.wait('manager', timeout=10.)
            if ok == ocs.OK:
                session.add_message('... manager Process has exited.')
            else:
                session.add_message('... timed-out waiting for manager Process exit!')

        # Schedule program exit.
        reactor.callLater(1., reactor.stop)

        return True, 'This HostManager should terminate in about 1 second.'


def _clsname_tool(name, new_suffix=None):
    try:
        i = name.index('[')
    except ValueError:
        i = len(name)
    base, suffix = name[:i], name[i:]
    if new_suffix is None:
        return base, suffix
    return base + new_suffix


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--initial-state', default=None,
                        choices=['up', 'down'],
                        help="Force a single target state for all agents, "
                        "on start-up.")
    pgroup.add_argument('--docker-compose', default=None,
                        help="Comma-separated list of docker-compose files "
                        "to parse and manage.")
    pgroup.add_argument('--docker-service-prefix', default='ocs-',
                        help="Prefix, to be used in combination with "
                        "instance-id, for recognizing docker services "
                        "that correspond to entries in site config.")
    pgroup.add_argument('--docker-compose-bin', default=None,
                        help="Path to docker-compose binary.  This "
                        "will be interpreted as a path relative to "
                        "current working directory.  If not specified, "
                        "will try to use `which docker-compose`.")
    pgroup.add_argument('--quiet', action='store_true',
                        help="Suppress output to stdout/stderr.")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='HostManager',
                                  parser=parser,
                                  args=args)

    if args.quiet:
        # For launch-to-background, disconnect stdio.
        null = os.open(os.devnull, os.O_RDWR)
        for stream in [sys.stdin, sys.stdout, sys.stderr]:
            os.dup2(null, stream.fileno())
        os.close(null)

    agent, runner = ocs_agent.init_site_agent(args)

    docker_composes = []
    if args.docker_compose:
        docker_composes = args.docker_compose.split(',')

    host_manager = HostManager(agent, docker_composes=docker_composes,
                               docker_service_prefix=args.docker_service_prefix)

    startup_params = {}
    if args.initial_state:
        startup_params = {'requests': [('all', args.initial_state)]}

    agent.register_process('manager',
                           host_manager.manager,
                           host_manager._stop_manager,
                           blocking=False,
                           startup=startup_params)
    agent.register_task('update', host_manager.update, blocking=False)
    agent.register_task('die', host_manager.die, blocking=False)

    reactor.addSystemEventTrigger('before', 'shutdown', agent._stop_all_running_sessions)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import time
import txaio

from datetime import datetime, timezone
from influxdb import InfluxDBClient
from influxdb.exceptions import InfluxDBClientError, InfluxDBServerError
from requests.exceptions import ConnectionError as RequestsConnectionError

# For logging
txaio.use_twisted()
LOG = txaio.make_logger()


def timestamp2influxtime(time, protocol):
    """Convert timestamp for influx, always in UTC.

    Args:
        time:
            ctime timestamp
        protocol:
            'json' or line'

    """
    if protocol == 'json':
        t_dt = datetime.fromtimestamp(time)
        # InfluxDB expects UTC by default
        t_dt = t_dt.astimezone(tz=timezone.utc)
        influx_t = t_dt.strftime("%Y-%m-%dT%H:%M:%S.%f")
    elif protocol == 'line':
        influx_t = int(time * 1e9)  # ns
    return influx_t


class Publisher:
    """
    Data publisher. This manages data to be published to the InfluxDB.

    This class should only be accessed by a single thread. Data can be passed
    to it by appending it to the referenced `incoming_data` queue.

    Args:
        incoming_data (queue.Queue):
            A thread-safe queue of (data, feed) pairs.
        host (str):
            host for InfluxDB instance.
        database (str):
            database name within InfluxDB to publish to
        port (int, optional):
            port for InfluxDB instance, defaults to 8086.
        protocol (str, optional):
            Protocol for writing data. Either 'line' or 'json'.
        gzip (bool, optional):
            compress influxdb requsts with gzip
        operate_callback (callable, optional):
            Function to call to see if failed connections should be
            retried (to prevent a thread from locking).

    Attributes:
        host (str):
            host for InfluxDB instance.
        port (int, optional):
            port for InfluxDB instance, defaults to 8086.
        db (str):
            database name within InfluxDB to publish to (from database arg)
        incoming_data:
            data to be published
        client:
            InfluxDB client connection

    """

    def __init__(self, host, database, incoming_data, port=8086, protocol='line',
                 gzip=False, operate_callback=None):
        self.host = host
        self.port = port
        self.db = database
        self.incoming_data = incoming_data
        self.protocol = protocol
        self.gzip = gzip

        print(f"gzip encoding enabled: {gzip}")
        print(f"data protocol: {protocol}")

        self.client = InfluxDBClient(host=self.host, port=self.port, gzip=gzip)

        db_list = None
        # ConnectionError here is indicative of InfluxDB being down
        while db_list is None:
            try:
                db_list = self.client.get_list_database()
            except RequestsConnectionError:
                LOG.error("Connection error, attempting to reconnect to DB.")
                self.client = InfluxDBClient(host=self.host, port=self.port, gzip=gzip)
                time.sleep(1)
            if operate_callback and not operate_callback():
                break

        db_names = [x['name'] for x in db_list]

        if self.db not in db_names:
            print(f"{self.db} DB doesn't exist, creating DB")
            self.client.create_database(self.db)

        self.client.switch_database(self.db)

    def process_incoming_data(self):
        """
        Takes all data from the incoming_data queue, and writes them to the
        InfluxDB.
        """
        payload = []
        LOG.debug("Pulling data from queue.")
        while not self.incoming_data.empty():
            data, feed = self.incoming_data.get()
            if feed['agg_params'].get('exclude_influx', False):
                continue

            # Formatted for writing to InfluxDB
            payload.extend(self.format_data(data, feed, protocol=self.protocol))

        # Skip trying to write if payload is empty
        if not payload:
            return

        try:
            LOG.debug("payload: {p}", p=payload)
            self.client.write_points(payload,
                                     batch_size=10000,
                                     protocol=self.protocol,
                                     )
            LOG.debug("wrote payload to influx")
        except RequestsConnectionError:
            LOG.error("InfluxDB unavailable, attempting to reconnect.")
            self.client = InfluxDBClient(host=self.host, port=self.port, gzip=self.gzip)
            self.client.switch_database(self.db)
        except InfluxDBClientError as err:
            LOG.error("InfluxDB Client Error: {e}", e=err)
        except InfluxDBServerError as err:
            LOG.error("InfluxDB Server Error: {e}", e=err)

    @staticmethod
    def _format_field_line(field_key, field_value):
        """Format key-value pair for InfluxDB line protocol."""
        # Strings must be in quotes for line protocol
        if isinstance(field_value, str):
            line = f'{field_key}="{field_value}"'
        else:
            line = f"{field_key}={field_value}"
        # Don't append 'i' to bool, which is a subclass of int
        if isinstance(field_value, int) and not isinstance(field_value, bool):
            line += "i"
        return line

    @staticmethod
    def format_data(data, feed, protocol):
        """Format the data from an OCS feed into a dict for pushing to InfluxDB.

        The scheme here is as follows:
            - agent_address is the "measurement" (conceptually like an SQL
              table)
            - feed names are an indexed "tag" on the data structure
              (effectively a table column)
            - keys within an OCS block's 'data' dictionary are the field names
              (effectively a table column)

        Args:
            data (dict):
                data from the OCS Feed subscription
            feed (dict):
                feed from the OCS Feed subscription, contains feed information
                used to structure our influxdb query
            protocol (str):
                Protocol for writing data. Either 'line' or 'json'.

        Returns:
            list: Data ready to publish to influxdb, in the specified protocol.

        """
        measurement = feed['agent_address']
        feed_tag = feed['feed_name']

        json_body = []

        # Reshape data for query
        for bk, bv in data.items():
            grouped_data_points = []
            times = bv['timestamps']
            num_points = len(bv['timestamps'])
            for i in range(num_points):
                grouped_dict = {}
                for data_key, data_value in bv['data'].items():
                    grouped_dict[data_key] = data_value[i]
                grouped_data_points.append(grouped_dict)

            for fields, time_ in zip(grouped_data_points, times):
                if protocol == 'line':
                    fields_line = []
                    for mk, mv in fields.items():
                        f_line = Publisher._format_field_line(mk, mv)
                        fields_line.append(f_line)

                    measurement_line = ','.join(fields_line)
                    t_line = timestamp2influxtime(time_, protocol='line')
                    line = f"{measurement},feed={feed_tag} {measurement_line} {t_line}"
                    json_body.append(line)
                elif protocol == 'json':
                    json_body.append(
                        {
                            "measurement": measurement,
                            "time": timestamp2influxtime(time_, protocol='json'),
                            "fields": fields,
                            "tags": {
                                "feed": feed_tag
                            }
                        }
                    )
                else:
                    LOG.warn(f"Protocol '{protocol}' not supported.")

        return json_body

    def run(self):
        """Main run iterator for the publisher. This processes all incoming
        data, removes stale providers, and writes active providers to disk.

        """
        self.process_incoming_data()

    def close(self):
        """Flushes all remaining data and closes InfluxDB connection."""
        pass



package_name = 'ocs'
agents = {
    'RegistryAgent': {'module': 'ocs.agents.registry.agent', 'entry_point': 'main'},
    'AggregatorAgent': {'module': 'ocs.agents.aggregator.agent', 'entry_point': 'main'},
    'HostManager': {'module': 'ocs.agents.host_manager.agent', 'entry_point': 'main'},
    'FakeDataAgent': {'module': 'ocs.agents.fake_data.agent', 'entry_point': 'main'},
    'InfluxDBAgent': {'module': 'ocs.agents.influxdb_publisher.agent', 'entry_point': 'main'},
    'BarebonesAgent': {'module': 'ocs.agents.barebones.agent', 'entry_point': 'main'},
}



import ocs

import txaio
txaio.use_twisted()

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.internet.error import ReactorNotRunning

from autobahn.wamp.types import ComponentConfig
from autobahn.twisted.wamp import ApplicationSession, ApplicationRunner
import deprecation


@deprecation.deprecated(
    deprecated_in='v0.6.1',
    details="Renamed to run_control_script"
)
def run_control_script2(*args, **kwargs):
    run_control_script(*args, **kwargs)


def run_control_script(function, parser=None, *args, **kwargs):
    """
    Run a function in a ControlClientSession, within a site_config
    paradigm, assuming that additional configuration will be passed
    through command line arguments.

    Args:
        function: The function to call.

        parser: argparse.ArgumentParser, with control script options
            pre-loaded.  If None, this will be created internally in
            the usual way.

    The command line is parsed and site configuration information is
    loaded.  The WAMP server information is used to configure the
    Control Client's connection before launching the function.  The
    function is invoked as::

        function(app, parser_args, *args, **kwargs)

    where app is the ControlClientSession and parser_args is the
    argparse.Namespace object including processing done by
    ocs.site_config.  Note that the observatory root_address is
    contained in parser_args.root_address.  Any additional arguments
    defined in the parser will also be present.

    This function can be invoked with parser=None, or else with a
    parser that has been initialized for control client purposes, like
    this::

        from ocs import client_t, site_config
        parser = site_control.add_arguments()  # initialized ArgParser
        parser.add_option('--target')          # Options for this client
        client_t.run_control_script2(my_script, parser=parser)

    In the my_script function, use parser_args.target to get the
    target.
    """
    if parser is None:
        parser = ocs.site_config.add_arguments()
    pargs = parser.parse_args()
    ocs.site_config.reparse_args(pargs, '*control*')
    server, realm = pargs.site_hub, pargs.site_realm
    session = ControlClientSession(ComponentConfig(realm, {}), function,
                                   [pargs] + list(args), kwargs)
    runner = ApplicationRunner(server, realm)
    runner.run(session, auto_reconnect=True)


class ControlClientSession(ApplicationSession):

    def __init__(self, config, script, script_args, script_kwargs):
        ApplicationSession.__init__(self, config)
        self.script = (script, script_args, script_kwargs)

    def onConnect(self):
        self.log.info('transport connected')
        self.join(self.config.realm)

    def onChallenge(self, challenge):
        self.log.info('authentication challenge received')

    # The Operations API...

    def onLeave(self, details):
        self.log.info('session left: {}'.format(details))
        self.disconnect()

    def onDisconnect(self):
        self.log.info('transport disconnected')
        try:
            reactor.stop()
        except ReactorNotRunning:
            pass

    @inlineCallbacks
    def onJoin(self, details):
        self.log.info('session joined: {}'.format(details))
        script, a, kw = self.script
        yield from script(self, *a, **kw)
        yield self.leave()


class OperationClient:
    """The endpoint client is associated with a single operation rather
    than with an operation server."""

    def __init__(self, app, root_address, op_name):
        self.app = app
        self.root = root_address
        self.op_name = op_name

    def request(self, action, params=None, timeout=None):
        return self.app.call(self.root + '.ops', action, self.op_name,
                             params=params, timeout=timeout)

    def status(self, params=None):
        return self.request('status', params=params)

    def start(self, params=None):
        return self.request('start', params=params)

    def wait(self, params=None, timeout=None):
        return self.request('wait', params=params, timeout=timeout)


class TaskClient(OperationClient):
    def abort(self, params=None):
        return self.request('abort', params=params)


class ProcessClient(OperationClient):
    def stop(self, params=None):
        return self.request('stop', params=params)



import os
import time
import yaml

from twisted.internet import reactor, utils, protocol
from twisted.internet.defer import inlineCallbacks


class ManagedInstance(dict):
    """Track properties of a managed Agent-instance.  This is just a dict
    with a schema docstring and an "init" function to set defaults.

    Properties that must be set explicitly by user:

    - 'management' (str): Either 'host', 'docker', or 'retired'.
    - 'agent_class' (str): The agent class name, which may include a
      suffix ([d] or [d?]) if the agent is managed through Docker.
      For instances corresponding to docker services that do not have
      a corresponding SCF entry, the value here will be '[docker]'.
    - 'instance_id' (str): The agent instance-id, or the docker
      service name if the instance is an unmatched docker service.
    - 'full_name' (str): agent_class:instance_id

    Properties that are given a default value by init function:

    - 'operable' (bool): indicates whether the instance can be
      manipulated (whether calls to up/down should be expected to
      work).
    - 'agent_script' (str): Path to the launcher script (if host
      system managed).  If docker-managed, this is the service name.
    - 'prot': The twisted ProcessProtocol object (if host system
      managed), or the DockerContainerHelper (if a docker container).
    - 'target_state' (state): The state we're trying to achieve (up or
      down).
    - 'next_action' (state): The thing HostManager needs to do next;
      this will sometimes indicate the "current state" (up or down),
      but sometimes it will carry a transitional state, such as
      "wait_start".
    - 'at' (float): a unix timestamp for transitional states
      (e.g. used to set how long to wait for something).
    - 'fail_times' (list of floats): unix timestamps when the instance
      process has stopped unexpectedly (used to identify "unstable"
      agents).

    """
    @classmethod
    def init(cls, **kwargs):
        # Note some core things are not included.
        self = cls({
            'agent_script': None,
            'operable': False,
            'prot': None,
            'next_action': 'down',
            'target_state': 'down',
            'fail_times': [],
            'at': 0,
        })
        self.update(kwargs)
        return self


def resolve_child_state(db):
    """Args:

      db (ManagedInstance): the instance state information.  This will
        be modified in place.

    Returns:

      Dict with important actions for caller to take.  Content is:

      - 'messages' (list of str): messages for the session.
      - 'launch' (bool): whether to launch a new instance.
      - 'terminate' (bool): whether to terminate the instance.
      - 'sleep' (float): maximum delay before checking back, or None
        if this machine doesn't care.

    """
    actions = {
        'launch': False,
        'terminate': False,
        'sleep': None,
    }

    messages = []
    sleeps = []

    # State machine.
    prot = db['prot']

    # If the entry is not "operable", send next_action to '?' and
    # don't try to do anything else.

    if not db['operable']:
        db['next_action'] = '?'

    # The uninterruptible transition state(s) are most easily handled
    # in the same way regardless of target state.

    # Transitional: wait_start, which bridges from start -> up.
    elif db['next_action'] == 'wait_start':
        if prot is not None:
            messages.append('Launched {full_name}'.format(**db))
            db['next_action'] = 'up'
        else:
            if time.time() >= db['at']:
                messages.append('Launch not detected for '
                                '{full_name}!  Will retry.'.format(**db))
                db['next_action'] = 'start_at'
                db['at'] = time.time() + 5.

    # Transitional: wait_dead, which bridges from kill -> idle.
    elif db['next_action'] == 'wait_dead':
        if prot is None:
            stat, t = 0, None
        else:
            stat, t = prot.status
        if stat is not None:
            db['next_action'] = 'down'
        elif time.time() >= db['at']:
            if stat is None:
                messages.append('Agent instance {full_name} '
                                'refused to die.'.format(**db))
                db['next_action'] = 'down'
        else:
            sleeps.append(db['at'] - time.time())

    # State handling when target is to be 'up'.
    elif db['target_state'] == 'up':
        if db['next_action'] == 'start_at':
            if time.time() >= db['at']:
                db['next_action'] = 'start'
            else:
                sleeps.append(db['at'] - time.time())
        elif db['next_action'] == 'start':
            messages.append(
                'Requested launch for {full_name}'.format(**db))
            actions['launch'] = True
            db['next_action'] = 'wait_start'
            now = time.time()
            db['at'] = now + 1.
        elif db['next_action'] == 'up':
            if prot is None:
                stat, t = 0, None
            else:
                stat, t = prot.status
            if stat is not None:
                messages.append('Detected exit of {full_name} '
                                'with code {stat}.'.format(stat=stat, **db))
                if hasattr(prot, 'lines'):
                    note = ''
                    lines = prot.lines['stderr']
                    if len(lines) > 50:
                        note = ' (trimmed)'
                        lines = lines[-20:]
                    messages.append('stderr output from {full_name}{note}: {}'
                                    .format('\n'.join(lines), note=note, **db))
                db['next_action'] = 'start_at'
                db['at'] = time.time() + 3
                db['fail_times'].append(time.time())
        else:  # 'down'
            db['next_action'] = 'start'

    # State handling when target is to be 'down'.
    elif db['target_state'] == 'down':
        if db['next_action'] == 'down':
            # The lines below will prevent HostManager from killing
            # Agents that suddenly seem to be alive.  With these
            # lines commented out, someone running "up" on a managed
            # docker-compose.yaml will see their Agents immediately
            # be brought down by HostManager.
            # if prot is not None and prot.status[0] is None:
            #    messages.append('Detected unexpected session for {full_name} '
            #                    '(probably docker); changing target state to "up".'.format(**db))
            #    db['target_state'] = 'up'

            # In fully managed mode, force a termination.
            if prot is not None and prot.status[0] is None:
                messages.append('Detected unexpected session for {full_name} '
                                '(probably docker); it will be shut down.'.format(**db))
                db['next_action'] = 'up'
        elif db['next_action'] == 'up':
            messages.append('Requesting termination of '
                            '{full_name}'.format(**db))
            actions['terminate'] = True
            db['next_action'] = 'wait_dead'
            db['at'] = time.time() + 5
        else:  # 'start_at', 'start'
            messages.append('Modifying state of {full_name} from '
                            '{next_action} to idle'.format(**db))
            db['next_action'] = 'down'

    # Should not get here.
    else:
        messages.append(
            'State machine failure: state={next_action}, target_state'
            '={target_state}'.format(**db))

    actions['messages'] = messages
    if len(sleeps):
        actions['sleep'] = min(sleeps)
    return actions


def stability_factor(times, window=120):
    """Given an increasing list of failure times, quantify the stability
    of the activity.

    A single failure, 10 seconds in the past, has a stability factor
    of 0.5; if there were additional failures before that, the
    stability factor will be lower.

    Returns a culled list of stop times and a stability factor (0 -
    1).

    """
    now = time.time()
    if len(times) == 0:
        return times, 1.
    # Only keep the last few failures, within our time window.
    times = [t for t in times[-200:-1]
             if t >= now - window] + times[-1:]
    dt = [5. / (now - t) for t in times]
    return times, max(1 - sum(dt), 0.)


class AgentProcessHelper(protocol.ProcessProtocol):
    def __init__(self, instance_id, cmd):
        super().__init__()
        self.status = None, None
        self.killed = False
        self.instance_id = instance_id
        self.cmd = cmd
        self.lines = {'stderr': [],
                      'stdout': []}

    def up(self):
        reactor.spawnProcess(self, self.cmd[0], self.cmd[:], env=os.environ)

    def down(self):
        self.killed = True
        # race condition, but it could be worse.
        if self.status[0] is None:
            reactor.callFromThread(self.transport.signalProcess, 'INT')

    # See https://twistedmatrix.com/documents/current/core/howto/process.html
    #
    # These notes, and the useless prototypes below them, are to get
    # us started when we come back here later to feed the process
    # output to high level logging somehow.
    #
    # In a successful launch, we see:
    # - connectionMade (at which point we closeStdin)
    # - inConnectionLost (which is then expected)
    # - childDataReceived(counter, message), output from the script.
    # - later, when process exits: processExited(status).  Status is some
    #   kind of object that knows the return code...
    # In a failed launch, it's the same except note that:
    # - The childDataReceived message contains the python traceback, on,
    #   e.g. realm error.  +1 - Informative.
    # - The processExited(status) knows the return code was not 0.
    #
    # Note that you implement childDataReceived instead of
    # "outReceived" and "errReceived".
    def connectionMade(self):
        self.transport.closeStdin()

    def inConnectionLost(self):
        pass

    def processExited(self, status):
        # print('%s.status:' % self.instance_id, status)
        self.status = status, time.time()

    def outReceived(self, data):
        self.lines['stdout'].extend(data.decode('utf8').split('\n'))
        if len(self.lines['stdout']) > 100:
            self.lines['stdout'] = self.lines['stdout'][-100:]

    def errReceived(self, data):
        self.lines['stderr'].extend(data.decode('utf8').split('\n'))
        if len(self.lines['stderr']) > 100:
            self.lines['stderr'] = self.lines['stderr'][-100:]


def _run_docker(args):
    return utils.getProcessOutputAndValue(
        'docker', args, env=os.environ)


class DockerContainerHelper:
    """Class for managing the docker container associated with some
    service.  Provides some of the same interface as
    AgentProcessHelper.  Pass in a service description dict (such as
    the ones returned by parse_docker_state).

    """

    def __init__(self, service, docker_bin=None):
        self.service = {}
        self.status = -1, time.time()
        self.killed = False
        self.instance_id = service['service']
        self.d = None
        self.update(service)

    def update(self, service):
        """Update self.status based on service info (in format returned by
        parse_docker_state).

        """
        self.service.update(service)
        if service['running']:
            self.status = None, time.time()
        else:
            self.status = service['exit_code'], time.time()
            self.killed = False

    def up(self):
        self.d = _run_docker(
            ['compose', '-f', self.service['compose_file'],
             'up', '-d', self.service['service']])
        self.status = None, time.time()

    def down(self):
        self.d = _run_docker(
            ['compose', '-f', self.service['compose_file'],
             'rm', '--stop', '--force', self.service['service']])
        self.killed = True


@inlineCallbacks
def parse_docker_state(docker_compose_file):
    """Analyze a docker compose.yaml file to get a list of services.
    Using docker compose ps and docker inspect, determine whether each
    service is running or not.

    Returns:
      A dict where the key is the service name and each value is a
      dict with the following entries:

      - 'compose_file': the path to the docker compose file
      - 'service': service name
      - 'container_found': bool, indicates whether a container for
        this service was found (whether or not it was running).
      - 'running': bool, indicating that a container for this service
        is currently in state "Running".
      - 'exit_code': int, which is either extracted from the docker
        inspect output or is set to 127.  (This should never be None.)

    """

    summary = {}

    compose = yaml.safe_load(open(docker_compose_file, 'r'))
    for key, cfg in compose.get('services', []).items():
        summary[key] = {
            'service': key,
            'running': False,
            'exit_code': 127,
            'container_found': False,
            'compose_file': docker_compose_file,
        }

    # Query docker compose for container ids...
    out, err, code = yield _run_docker(
        ['compose', '-f', docker_compose_file, 'ps', '-q'])
    if code != 0:
        raise RuntimeError("Could not run docker compose or could not parse "
                           "compose.yaml file; exit code %i, error text: %s" %
                           (code, err))

    cont_ids = [line.strip() for line in out.decode('utf8').split('\n')
                if line.strip() != '']

    # Run docker inspect.
    for cont_id in cont_ids:
        try:
            info = yield _inspectContainer(cont_id, docker_compose_file)
        except RuntimeError as e:
            print(f'Warning, failed to inspect container {cont_id}; {e}.')
            continue
        if info is None:
            continue

        service = info.pop('service')
        if service not in summary:
            raise RuntimeError("Consistency problem: image does not self-report "
                               "as a listed service? (%s)" % (service))
        summary[service].update(info)

    return summary


@inlineCallbacks
def _inspectContainer(cont_id, docker_compose_file):
    """Run docker inspect on cont_id, return dict with the results."""
    out, err, code = yield _run_docker(
        ['inspect', cont_id])
    if code != 0 and 'No such object' in err.decode('utf8'):
        # This is likely due to a race condition where some
        # container was brought down since we ran docker compose.
        # Return None to indicate this -- caller should just ignore for now.
        print(f'(_inspectContainer: warning, no such object: {cont_id}')
        return None
    elif code != 0:
        raise RuntimeError(
            f'Trouble running "docker inspect {cont_id}".\n'
            f'stdout: {out}\n  stderr {err}')
    # Reconcile config against docker compose ...
    info = yaml.safe_load(out)[0]
    config = info['Config']['Labels']
    _dc_file = os.path.join(config['com.docker.compose.project.working_dir'],
                            config['com.docker.compose.project.config_files'])
    if not os.path.samefile(docker_compose_file, _dc_file):
        raise RuntimeError("Consistency problem: container started from "
                           "some other compose file?\n%s\n%s" % (docker_compose_file, _dc_file))
    service = config['com.docker.compose.service']
    return {
        'service': service,
        'running': info['State']['Running'],
        'exit_code': info['State'].get('ExitCode', 127),
        'container_found': True,
    }



import os
import binascii
import time

from typing import Dict

import txaio
txaio.use_twisted()

from ocs.ocs_feed import Block, Feed

import so3g
from spt3g import core


HKAGG_VERSION = 2
_g3_casts = {
    str: core.G3String,
    int: core.G3Int,
    float: core.G3Double,
    bool: core.G3Bool,
}
_g3_list_casts = {
    str: core.G3VectorString,
    int: core.G3VectorInt,
    float: core.G3VectorDouble,
    bool: core.G3VectorBool,
}

LOG = txaio.make_logger()


def g3_cast(data, time=False):
    """
    Casts a generic datatype into a corresponding G3 type. With:
        int   -> G3Int
        str   -> G3String
        float -> G3Double
        bool  -> G3Bool

    and lists of type X will go to G3VectorX. If ``time`` is set to True, will
    convert to G3Time or G3VectorTime with the assumption that ``data`` consists
    of unix timestamps.

    Args:
        data (int, str, float, or list):
            Generic data to be converted to a corresponding G3Type.
        time (bool, optional):
            If True, will assume data contains unix timestamps and try to cast
            to G3Time or G3VectorTime.

    Returns:
        g3_data:
            Corresponding G3 datatype.
    """
    is_list = isinstance(data, list)
    if is_list:
        dtype = type(data[0])
        if not all(isinstance(d, dtype) for d in data):
            raise TypeError("Data list contains varying types!")
    else:
        dtype = type(data)
    if dtype not in _g3_casts.keys():
        raise TypeError("g3_cast does not support type {}. Type must "
                        "be one of {}".format(dtype, _g3_casts.keys()))
    if is_list:
        if time:
            return core.G3VectorTime(list(map(
                lambda t: core.G3Time(t * core.G3Units.s), data)))
        else:
            cast = _g3_list_casts[type(data[0])]
            return cast(data)
    else:
        if time:
            return core.G3Time(data * core.G3Units.s)
        else:
            cast = _g3_casts[type(data)]
            return cast(data)


def generate_id(hksess):
    """
    Generates a unique session id based on the start_time, process_id,
    and hksess description.

    Args:
        hksess (so3g.HKSessionHelper)
    """
    # Maybe this should go directly into HKSessionHelper
    elements = [
        (int(hksess.start_time), 32),
        (os.getpid(), 14),
        (binascii.crc32(bytes(hksess.description, 'utf8')), 14)
    ]
    agg_session_id = 0
    for i, b in elements:
        agg_session_id = (agg_session_id << b) | (i % (1 << b))
    return agg_session_id


def make_filename(base_dir, make_subdirs=True):
    """
    Creates a new filename based on the time and base_dir.
    If make_subdirs is True, all subdirectories will be automatically created.
    I don't think there's any reason that this shouldn't be true...

    Args:
        base_dir (path):
            Base path where data should be written.
        make_subdirs (bool):
            True if func should automatically create non-existing subdirs.
    """
    start_time = time.time()

    subdir = os.path.join(base_dir, "{:.5}".format(str(start_time)))

    if not os.path.exists(subdir):
        if make_subdirs:
            try:
                os.makedirs(subdir)
            except PermissionError as e:
                LOG.error("Permission error encountered while trying to create "
                          f"data sub-directory: {e}")
                raise e
        else:
            raise FileNotFoundError("Subdir {} does not exist"
                                    .format(subdir))

    time_string = int(start_time)
    filename = os.path.join(subdir, "{}.g3".format(time_string))
    return filename


class Provider:
    """
    Stores data for a single provider (OCS Feed).
    This class should only be accessed via a single thread.

    Args:
        addresss (string):
            Full address of the provider
        sessid (string):
            session_id of the provider
        prov_id (bool):
            id assigned to the provider by the HKSessionHelper
        frame_length (float, optional):
            Time (in seconds) before data should be written into a frame. Defaults to 5 min.
        fresh_time (float, optional):
            Time (in seconds) before provider should be considered stale. Defaults to 3 min.

    Attributes:

        blocks (dict):
            All blocks that are written by provider.
        frame_start_time (float):
            Start time of current frame
        fresh_time (float):
            time (in seconds) that the provider can go without data before being
            labeled stale, and scheduled to be removed
        last_refresh (time):
            Time when the provider was last refreshed (either through data or
            agent heartbeat).
        last_block_received (str):
            String of the last block_name received.

        log (txaio.Logger):
            txaio logger

    """

    def __init__(self, address, sessid, prov_id, frame_length=5 * 60, fresh_time=3 * 60):
        self.address = address
        self.sessid = sessid
        self.frame_length = frame_length
        self.prov_id = prov_id
        self.log = txaio.make_logger()

        self.blocks = {}

        # When set to True, provider will be written and removed next agg cycle
        self.frame_start_time = None

        self.fresh_time = fresh_time
        self.last_refresh = time.time()  # Determines if
        self.last_block_received = None

    def encoded(self):
        return {
            'last_refresh': self.last_refresh,
            'sessid': self.sessid,
            'stale': self.stale(),
            'last_block_received': self.last_block_received
        }

    def refresh(self):
        """Refresh provider"""
        self.last_refresh = time.time()

    def stale(self):
        """Returns true if provider is stale and should be removed"""
        return (time.time() - self.last_refresh) > self.fresh_time

    def new_frame_time(self):
        """Returns true if its time for a new frame to be written"""
        if self.frame_start_time is None:
            return False

        return (time.time() - self.frame_start_time) > self.frame_length

    def empty(self):
        """Returns true if all blocks are empty"""
        for _, b in self.blocks.items():
            if not b.empty():
                return False

        return True

    def _verify_provider_data(self, data):
        """Check the provider data for invalid field names. Meant to be used in
        combination with Provider._rebuild_invalid_data().

        Args:
            data (dict): data dictionary passed to Provider.save_to_block()

        Returns:
            bool: True if all field names valid. False if any invalid names found

        """
        verified = True
        for block_name, block_dict in data.items():
            for field_name, field_values in block_dict['data'].items():
                try:
                    Feed.verify_data_field_string(field_name)
                except ValueError:
                    self.log.error("data field name '{field}' is "
                                   + "invalid, removing invalid characters.",
                                   field=field_name)
                    verified = False

        return verified

    @staticmethod
    def _check_for_duplicate_names(field_name, name_list):
        """Check name_list for matching field names and modify field_name if
        matches are found.

        The results of ocs_feed.Feed.enforce_field_name_rules() are not guarenteed
        to be unique. This method will check field_name against a list of
        existing field names and try to append '_N', with N being a zero padded
        integer up to 99. Longer integers, though not expected to see use, are
        also supported, though will not be zero padded.

        In the event the field name is at the maximum allowed length, we remove
        some characters before appending the additional underscore and integer.

        Examples:
            >>> current_field_names = ['test', 'test_01']
            >>> name = 'test'
            >>> new_name = Provider._check_for_duplicate_names(name, current_field_names)
            >>> print(new_name)
            test_02

        Args:
            field_name (str): field name to check against name_lsit
            name_list (list): list of field names already in a Block

        Returns:
            str: A new field name that is not already in name_list

        """
        name_index = 1

        while field_name in name_list:
            suffix = f'_{name_index:02}'
            suf_len = len(suffix)
            field_name = field_name[:255 - suf_len] + suffix
            name_index += 1

        return field_name

    def _rebuild_invalid_data(self, data):
        """Rebuild an invalid data dictionary.

        Args:
            data (dict): data dictionary passed to Provider.save_to_block().

        Returns:
            dict: A rebuilt data dictionary with invalid characters stripped
                  from the field names, limited to 255 characters in length.

        """
        new_data = {}
        for block_name, block_dict in data.items():
            new_data[block_name] = {}
            # rebuild block_dict
            for k, v in block_dict.items():
                if k == 'data':
                    new_data[block_name]['data'] = {}
                    new_field_names = []
                    for field_name, field_values in block_dict['data'].items():
                        new_field_name = Feed.enforce_field_name_rules(field_name)

                        # Catch instance where rule enforcement strips all characters
                        if not new_field_name:
                            new_field_name = Feed.enforce_field_name_rules("invalid_field_" + field_name)

                        new_field_name = Provider._check_for_duplicate_names(new_field_name,
                                                                             new_field_names)

                        new_data[block_name]['data'][new_field_name] = field_values

                        new_field_names.append(new_field_name)
                else:
                    new_data[block_name][k] = v

        return new_data

    def save_to_block(self, data):
        """Saves a list of data points into blocks. A block will be created
        for any new block_name.

        Examples:
            The format of data is shown in the following example:

            >>> data = {'test': {'block_name': 'test',
                             'timestamps': [time.time()],
                             'data': {'key1': [1],
                                      'key2': [2]},
                             }
                       }
            >>> prov.save_to_block(data)

            Note the block name shows up twice, once as the dict key in the
            outer data dictionary, and again under the 'block_name' value.
            These must match -- in this instance both the word 'test'.

        Args:
            data (dict): data dictionary from incoming data queue

        """
        self.refresh()

        if self.frame_start_time is None:
            # Get min frame time out of all blocks
            self.frame_start_time = time.time()
            for _, b in data.items():
                if b['timestamps']:
                    self.frame_start_time = min(self.frame_start_time, b['timestamps'][0])

        self.log.debug('data passed to block: {d}', d=data)
        verified = self._verify_provider_data(data)

        if not verified:
            self.log.info('rebuilding data containing invalid field name')
            data = self._rebuild_invalid_data(data)
            self.log.debug('data after rebuild: {d}', d=data)

        for key, block in data.items():
            try:
                b = self.blocks[key]
            except KeyError:
                self.blocks[key] = Block(
                    key, block['data'].keys(),
                )
                b = self.blocks[key]

            b.extend(block)
            self.last_block_received = key

    def clear(self):
        """Clears all blocks and resets the frame_start_time"""
        for _, b in self.blocks.items():
            b.clear()

        self.frame_start_time = None

    def to_frame(self, hksess=None, clear=False):
        """
        Returns a G3Frame based on the provider's blocks.

        Args:
            hksess (optional):
                If provided, the frame will be based off of hksession's data frame.
                If the data will be put into a clean frame.
            clear (bool):
                Clears provider data if True.
        """

        if hksess is not None:
            frame = hksess.data_frame(prov_id=self.prov_id)
        else:
            frame = core.G3Frame(core.G3FrameType.Housekeeping)

        frame['address'] = self.address
        frame['provider_session_id'] = self.sessid

        block_names = []
        for block_name, block in self.blocks.items():
            if not block.timestamps:
                continue
            try:
                m = core.G3TimesampleMap()
                m.times = g3_cast(block.timestamps, time=True)
                for key, ts in block.data.items():
                    m[key] = g3_cast(ts)
            except Exception as e:
                self.log.warn("Error received when casting timestream! {e}",
                              e=e)
                continue
            frame['blocks'].append(m)
            block_names.append(block_name)

        if 'block_names' in frame:
            frame['block_names'].extend(block_names)
        else:
            frame['block_names'] = core.G3VectorString(block_names)

        if clear:
            self.clear()
        return frame


class G3FileRotator(core.G3Module):
    """
    G3 module which handles file rotation.
    After time_per_file has elapsed, the rotator will end that file and create
    a new file with the `filename` function.
    It will write the last_session and last_status frame to any new file if they
    exist.

    This class should only be accessed via a single thread.

    Args:
        time_per_file (float):
            time (seconds) before a new file should be written
        filename (callable):
            function that generates new filenames.

    Attributes:
        filename (function):
            Function to call to create new filename on rotation
        file_start_time (int):
            Start time for current file
        writer (core.G3Writer):
            G3Writer object for current file. None if no file is open.
        last_session (core.G3Frame):
            Last session frame written to disk. This is stored and written as
            the first frame on file rotation.
        last_status (core.G3Frame):
            Last status frame written to disk. Stored and written as the second
            frame on file rotation.
        current_file (str, optional):
            Path to the current file being written.
    """

    def __init__(self, time_per_file, filename):
        self.time_per_file = time_per_file
        self.filename = filename
        self.log = txaio.make_logger()

        self.file_start_time = None
        self.writer = None
        self.last_session = None
        self.last_status = None
        self.current_file = None

    def close_file(self):
        if self.writer is not None:
            self.writer(core.G3Frame(core.G3FrameType.EndProcessing))
            self.writer = None

    def flush(self):
        """Flushes current g3 file to disk"""
        if self.writer is not None:
            self.writer.Flush()

    def Process(self, frames):
        """
        Writes frame to current file. If file has not been started
        or time_per_file has elapsed, file is closed and a new file is created
        by `filename` function passed to constructor
        """
        for frame in frames:
            ftype = frame['hkagg_type']

            if ftype == so3g.HKFrameType.session:
                self.last_session = frame
            elif ftype == so3g.HKFrameType.status:
                self.last_status = frame

            if self.writer is None:
                self.current_file = self.filename()
                self.log.info("Creating file: {}".format(self.current_file))
                self.writer = core.G3Writer(self.current_file)
                self.file_start_time = time.time()

                if ftype in [so3g.HKFrameType.data, so3g.HKFrameType.status]:
                    if self.last_session is not None:
                        self.writer(self.last_session)

                if ftype == so3g.HKFrameType.data:
                    if self.last_status is not None:
                        self.writer(self.last_status)

            self.writer(frame)

        if (time.time() - self.file_start_time) > self.time_per_file:
            self.close_file()

        return frames


class Aggregator:
    """Data aggregator. This manages a collection of providers, and contains
    methods to write them to disk.

    This class should only be accessed by a single thread. Data can be passed
    to it by appending it to the referenced `incoming_data` queue.

    Args:
        incoming_data (queue.Queue):
            A thread-safe queue of (data, feed) pairs.
        time_per_file (float):
            Time (sec) before a new file should be written to disk.
        data_dir (path):
            Base directory for new files to be written.
        session (OpSession, optional):
            Session object of current agent process. If not specified, session
            data will not be written.

    Attributes:
        log (txaio.Logger):
            txaio logger
        hksess (so3g.HKSessionHelper):
            HKSession helper that assigns provider id's to providers,
            and constructs so3g frames.
        writer (G3Module):
            Module to use to write frames to disk.
        providers (Dict[Provider]):
            dictionary of active providers, indexed by the hksess's assigned
            provider id.
        pids (Dict[Int]):
            Dictionary of provider id's assigned by the hksession. Indexed
            by (prov address, session_id).
        write_status (bool):
            If true, a status frame will be written next time providers are
            written to disk. This is set to True whenever a provider is added
            or removed.
    """

    def __init__(self, incoming_data, time_per_file, data_dir, session=None):
        self.log = txaio.make_logger()

        self.hksess = so3g.hk.HKSessionHelper(description="HK data",
                                              hkagg_version=HKAGG_VERSION)
        self.hksess.start_time = time.time()
        self.hksess.session_id = generate_id(self.hksess)

        self.incoming_data = incoming_data

        self.writer = G3FileRotator(
            time_per_file,
            lambda: make_filename(data_dir, make_subdirs=True),
        )
        self.writer.Process([self.hksess.session_frame()])

        self.providers: Dict[Provider] = {}  # by prov_id
        self.pids = {}  # By (address, sessid)
        self.provider_archive: Dict[Provider] = {}

        self.write_status = False
        self.session = session

    def process_incoming_data(self):
        """
        Takes all data from the incoming_data queue, and puts them into
        provider blocks.
        """
        while not self.incoming_data.empty():

            data, feed = self.incoming_data.get()
            agg_params = feed['agg_params']

            if agg_params.get('exclude_aggregator', False):
                continue

            address = feed['address']
            sessid = feed['session_id']

            pid = self.pids.get((address, sessid))
            if pid is None:
                prov_kwargs = {}
                for key in ['frame_length', 'fresh_time']:
                    if key in agg_params:
                        prov_kwargs[key] = agg_params[key]

                pid = self.add_provider(address, sessid, **prov_kwargs)

            prov = self.providers[pid]
            prov.save_to_block(data)

    def add_provider(self, prov_address, prov_sessid, **prov_kwargs):
        """
        Registers a new provider and writes a status frame.

        Args:
            prov_address (str):
                full address of provider
            prov_sessid (str):
                session id of provider

        Optional Arguments:
            Additional kwargs are passed directly to the Provider constructor,
            so defaults are set there.
        """
        pid = self.hksess.add_provider(description=prov_address)

        self.providers[pid] = Provider(
            prov_address, prov_sessid, pid, **prov_kwargs
        )
        self.provider_archive[prov_address] = self.providers[pid]

        self.log.info("Adding provider {}".format(prov_address))

        self.pids[(prov_address, prov_sessid)] = pid
        self.write_status = True
        return pid

    def remove_provider(self, prov):
        """
        Writes remaining provider data to frame and removes provider.

        Args:
            prov (Provider):
                provider object that should be removed.
        """
        pid = prov.prov_id
        addr, sessid = prov.address, prov.sessid

        if not prov.empty():
            self.writer.Process([prov.to_frame(self.hksess, clear=False)])

        self.log.info("Removing provider {}".format(prov.address))
        self.hksess.remove_provider(pid)
        del self.providers[pid]
        del self.pids[(addr, sessid)]
        self.write_status = True

    def remove_stale_providers(self):
        """
        Loops through all providers and check if they've gone stale. If they
        have, write their remaining data to disk (they shouldn't have any)
        and delete them.
        """
        stale_provs = []

        for pid, prov in self.providers.items():
            if prov.stale():
                self.log.info("Provider {} went stale".format(prov.address))
                stale_provs.append(prov)

        for prov in stale_provs:
            self.remove_provider(prov)

    def write_to_disk(self, clear=True, write_all=False):
        """
        Loop through all providers, and write their data to the frame_queue
        if they have surpassed their frame_time, or if write_all is True.

        Args:
            clear (bool):
                If True, provider data is cleared after write
            write_all (bool):
                If true all providers are written to disk regardless of whether
                frame_time has passed.
        """

        frames = []

        if self.write_status:
            frames.append(self.hksess.status_frame())
            self.write_status = False

        for pid, prov in self.providers.items():
            if prov.empty():
                continue
            if write_all or prov.new_frame_time():
                frames.append(prov.to_frame(self.hksess, clear=clear))

        self.writer.Process(frames)

    def run(self):
        """
        Main run iterator for the aggregator. This processes all incoming data,
        removes stale providers, and writes active providers to disk.
        """
        self.process_incoming_data()
        self.remove_stale_providers()
        self.write_to_disk()
        self.writer.flush()
        if self.session is not None:
            self.session.data = {
                'current_file': self.writer.current_file,
                'providers': {}
            }
            for addr, prov in self.provider_archive.items():
                self.session.data['providers'][addr] = prov.encoded()

    def close(self):
        """Flushes all remaining providers and closes file."""
        self.write_to_disk(write_all=True)
        self.writer.close_file()



import argparse
import sys
import code
import readline
import rlcompleter
import re

try:
    import IPython
    use_ipython = True
except ImportError:
    use_ipython = False

try:
    from twisted.internet import reactor, defer
    from autobahn.wamp.types import SubscribeOptions
    from autobahn.twisted.wamp import ApplicationSession, ApplicationRunner
    use_twisted = True
except ModuleNotFoundError:
    use_twisted = False

from ocs.ocs_client import OCSClient
from ocs import site_config, base

DESCRIPTION = """
This script provides a quick way to start a python or ipython shell to
interact with an OCS Agent.  You will need to set OCS_CONFIG_DIR
environment variable to the directory containing default.yaml, or else
use --site-* options to specify your configuration.

If you know the instance-id of the Agent you want to talk to, run::

  %(prog)s shell INSTANCE_ID

If you want to listen to heartbeat feeds to get a list of Agents in
the system, run::

  %(prog)s scan

"""


def get_parser():
    # What you put into description vs. usage vs. help is rather
    # subtle ... if making edits, confirm that "ocs-client-cli -h" and
    # "ocs-client-cli command -h" both render nicely.
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    client_sp = parser.add_subparsers(
        dest='command', metavar='command', help="Command (\"%(prog)s {command} -h\" for more help)")

    # shell
    p = client_sp.add_parser('shell', help="Start an interactive python session with "
                             "an OCSClient instantiated.")
    p.add_argument('instance_id', nargs='*', help="E.g. aggregator or fakedata-1.  Pass more than one to "
                   "get multiple clients.")
    p.add_argument('--simple', action='store_true', help="Do not use ipython for the shell, even if it is "
                   "available.")

    # scan
    p = client_sp.add_parser('scan', help="Gather and print list of Agents.")
    p.add_argument('--details', action='store_true', help="List all Operations with their current status OpCode.")
    p.add_argument('--use-registry', nargs='?', const='registry', help="Query the registry (faster than listening for heartbeats). "
                   "Pass the registry instance_id as an argument (default to 'registry').")

    # scan
    p = client_sp.add_parser('listen', help="Subscribe to feed(s) and dump to stdout.")
    p.add_argument('feed_selector', help="Feed name, which can include wildcard matching (double "
                   " ..).  E.g., try 'observatory..feeds.heartbeat'")

    return parser

# Note there's a similar function to this in ocsbow ... consider
# combining effort...


def decode_exception(args):
    """Decode exceptions from WAMP http interface."""
    try:
        text, data = args[0][4:6]
        assert (text.startswith('wamp.') or text.startswith('client_http.'))
    except Exception:
        return False, args, str(args)
    return True, text, str(data)


def get_instance_id(full_address, args):
    prefix = args.address_root + '.'
    assert (full_address.startswith(prefix))
    return full_address[len(prefix):]


def listen(parser, args):
    if not use_twisted:
        parser.error('The "listen" function requires twisted and autobahn packages.')
    feeds = args.feed_selector
    print(f'Subscribing to {feeds}')

    class Listener(ApplicationSession):
        @defer.inlineCallbacks
        def onJoin(self, details):
            topic = feeds
            options = SubscribeOptions(match='wildcard', details=True)
            yield self.subscribe(self.on_event, topic, options=options)

        def on_event(self, msg, details=None):
            print(f'[{details.topic}] {msg}')

        def onDisconnect(self):
            if reactor.running:
                reactor.stop()

    url = args.site_hub
    realm = args.site_realm
    runner = ApplicationRunner(url, realm)
    runner.run(Listener)


def scan(parser, args):
    if args.site_http is None:
        parser.error('Unable to find the OCS config; set OCS_CONFIG_DIR?')

    if args.use_registry:
        reg_addr = f'{args.address_root}.{args.use_registry}'
        try:
            c = OCSClient(get_instance_id(reg_addr, args), args=args)
        except RuntimeError as e:
            parsed, err_name, text = decode_exception(e.args)
            if parsed and err_name == 'wamp.error.no_such_procedure':
                parser.error(
                    f'Failed to connect to registry at {reg_addr}; the registry '
                    'must be running for "scan" to work.')
            else:
                raise e
        status = c.main.status()
        info = status.session['data']
        adjective = 'Registered'

    else:
        if not use_twisted:
            parser.error('The "scan" function requires twisted and autobahn '
                         'unless --use-registry is passed.')
        beats = {}

        class Listener(ApplicationSession):
            @defer.inlineCallbacks
            def onJoin(self, details):
                topic = f'{args.address_root}..feeds.heartbeat'
                options = SubscribeOptions(match='wildcard', details=True)
                yield self.subscribe(self.on_event, topic, options=options)

            def on_event(self, msg, details=None):
                beats[details.topic] = msg

            def onDisconnect(self):
                if reactor.running:
                    reactor.stop()

        url = args.site_hub
        realm = args.site_realm
        runner = ApplicationRunner(url, realm)

        print('Listening to heartbeat feeds for 2 seconds ...')
        reactor.callLater(2.0, reactor.stop)
        runner.run(Listener)
        # Un-log
        if hasattr(sys, '__stdout__'):
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__

        # Convert to resemble registry format.  Originally, heartbeat
        # data was just the integer 0; then with OpCode it became a
        # dict mapping op_name -> op_code value.
        info = {}
        for k, v in beats.items():
            instance_id = re.search(f'{args.address_root}.(.*).feeds.heartbeat', k)[1]
            if not isinstance(v[0], dict):
                v[0] = {'old_agent_no_opcodes': base.OpCode.EXPIRED}
            info[instance_id] = {'op_codes': v[0]}
        adjective = 'Detected'

    # Get agent list.
    print(f'List of {adjective} Agents: ({len(info)})')
    for addr, data in info.items():
        try:
            instance_id = get_instance_id(addr, args)
        except BaseException:
            instance_id = addr
        print(f'  {instance_id}')
        if args.details:
            for k, v in data['op_codes'].items():
                vs = base.OpCode(v).name
                print(f'    {k:20}: {vs}')
    print()


def shell(parser, args):
    if len(args.instance_id) == 0:
        parser.error('No instance_id provided.')

    cs = [OCSClient(iid, args=args) for iid in args.instance_id]
    vars = {'clients': cs}

    if len(cs) > 1:
        banner = 'ocs-client-cli --\n'
        for i, iid in enumerate(args.instance_id):
            banner += "  -- use clients[%i].<op_name> to talk to %s\n" % (i, iid)

    else:
        banner = "ocs-client-cli -- use client.<op_name> to talk to %s" % args.instance_id[0]
        vars['client'] = cs[0]

    if use_ipython and not args.simple:
        # IPython interpreter
        IPython.embed(header=banner, user_ns=vars, colors='linux')
    else:
        # Standard interpreter with tab-completion.
        readline.set_completer(rlcompleter.Completer(vars).complete)
        readline.parse_and_bind("tab: complete")
        code.InteractiveConsole(vars).interact(banner=banner)


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = get_parser()

    # Note this call adds a bunch of args to the parser, and parses them
    # including looking up the site config file and loading defaults from
    # there.
    args = site_config.parse_args(agent_class='*control*',
                                  parser=parser, args=args)

    if args.command == 'scan':
        scan(parser, args)
    elif args.command == 'shell':
        shell(parser, args)
    elif args.command == 'listen':
        listen(parser, args)
    else:
        parser.error(f"Unknown command '{args.command}'")



#!/usr/bin/env python3
# Brian Koopman

import os
import time

from progress.bar import Bar

from so3g import hk
from ocs.ocs_feed import Feed

from colorama import init, Fore, Style
init()


def _build_file_list(target):
    """Build list of files to scan.

    Parameters
    ----------
    target : str
        File or directory to scan.

    Returns
    -------
    list
        List of full paths to files for scanning.

    """
    _file_list = []
    if os.path.isfile(target):
        _file_list.append(target)
    elif os.path.isdir(target):
        a = os.walk(target)
        for root, _, _file in a:
            for g3 in _file:
                if g3[-2:] == "g3":
                    _file_list.append(os.path.join(root, g3))

    return _file_list


class DataChecker:
    """Check data for latest feeds and fields.

    Parameters
    ----------
    target : str
        File or directory to scan.
    verbose : bool
        Verbose output flag

    Attributes
    ----------
    hkas : so3g.hk.HKArchiveScanner
        HKArchiveScanner for reading in the data
    cat :
        Finalized HKArchiveScanner
    target : str
        File or directory to scan.
    verbose : bool
        Verbose output flag
    fields : dict
        fields returned from a cat.get_fields call
    timelines : dict
        timelines returned from a cat.get_fields call
    instances : dict
        Agent/feed/field information for each instance-id, format described in
        _populate_instances docstring

    """

    def __init__(self, target, verbose=False):
        self.target = target
        self.verbose = verbose

        self.hkas = hk.HKArchiveScanner()
        self.cat = None

        self.fields = None
        self.timelines = None

        self.instances = {}

        # Private attributes
        self._file_list = _build_file_list(target)
        self._field_count = 0

    def scan_files(self):
        """Scan all the files with the HKArchiveScanner for later processing."""
        _bar = Bar('Scanning', max=len(self._file_list))
        for _file in self._file_list:
            try:
                self.hkas.process_file(_file)
            except Exception as e:
                print(e)
            _bar.next()
        _bar.finish()

        self.cat = self.hkas.finalize()

    def _populate_instances(self):
        """Populate the instances dictionary with information about each agent,
        the agent's feeds, and the associated fields.

        For each instance ID there is a dictionary like:
        {INSTANCE_ID:
            {FEED1:
                {"fields":
                   {FIELD1:
                       {'full_name': 'ADDRESS_ROOT.INSTANCE_ID.feeds.FEED.FIELD',
                        't_last': float,
                        'v_last': float},
                    FIELD2:
                       {'full_name': 'ADDRESS_ROOT.INSTANCE_ID.feeds.FEED.FIELD',
                        't_last': float,
                        'v_last': float}
                    },
                 "t_last" : lowest t_last in all fields
                }
             FEED2: {},
             FEED3: {},
             ...
            }
        }

        """

        for field_name in sorted(self.fields):
            site, instance_id, _, feed, field = field_name.split(".")

            if instance_id not in self.instances:
                self.instances[instance_id] = {}

            if feed not in self.instances[instance_id]:
                self.instances[instance_id][feed] = {"fields": {},
                                                     "t_last": None}

            if field not in self.instances[instance_id][feed]:
                self.instances[instance_id][feed]["fields"][field] = \
                    {'full_name': field_name,
                     't_last': None,
                     'v_last': None}
                self._field_count += 1

    def process_files(self):
        """Process all files, resquesting simple data points for each dataset
        and determining the last time we saw the field and the last value that
        field held.

        """
        _bar = Bar('Processing', max=self._field_count)
        for instance_id, feeds in self.instances.items():
            for feed, fields in feeds.items():
                for field, d_info in fields['fields'].items():
                    t, x = self.cat.simple(d_info['full_name'])
                    d_info['t_last'] = t[-1]
                    d_info['v_last'] = x[-1]

                    if fields['t_last'] is None:
                        fields['t_last'] = t[-1]
                    elif t[-1] < fields['t_last']:
                        fields['t_last'] = t[-1]
                    else:
                        pass

                    _bar.next()
        _bar.finish()
        print()

    def run(self):
        """Run data checker, scan and process all files in target."""
        self.scan_files()
        self.fields, self.timelines = self.cat.get_fields()
        self._populate_instances()
        self.process_files()

    def __str__(self):
        """Print informative string representing the data checker results."""
        description_string = ""

        for instance_id, feeds in self.instances.items():
            description_string += f"{instance_id}\n"
            for feed, fields in feeds.items():
                if not self.verbose:
                    field_t_diff = time.time() - fields['t_last']
                    if field_t_diff > 600:
                        description_string += f"  {feed}: " + Fore.RED + f"{field_t_diff:.1f} s old\n" + Style.RESET_ALL
                    else:
                        description_string += f"  {feed}: {field_t_diff:.1f} s old\n"
                else:
                    # Determine width of "Field" column
                    field_str_len = 20
                    for _f in fields['fields']:
                        if len(_f) > field_str_len:
                            field_str_len = len(_f)

                    description_string += f"  {feed}\n"
                    # 20 per fixed field, 9 for dividers, and field_str_len
                    description_string += "  " + "-" * (69 + field_str_len) + "\n"

                    # Header string
                    desc_substring = "  "

                    # Field
                    _field_string = "Field".rjust(field_str_len)
                    desc_substring += _field_string + " | "

                    # Last Seen [s ago]
                    _t_diff_string = "{:>20}".format("Last Seen [s ago]")
                    desc_substring += _t_diff_string + " | "

                    # Seen At [ctime]
                    _t_last_string = "{:>20}".format("Seen At [ctime]")
                    desc_substring += _t_last_string + " | "

                    # Value
                    _v_last_string = "{:>20}".format("Value")
                    desc_substring += _v_last_string

                    description_string += desc_substring + "\n"

                    description_string += "  " + "-" * (69 + field_str_len) + "\n"

                    for field, d_info in fields['fields'].items():
                        t_diff = time.time() - d_info['t_last']
                        desc_substring = "  "

                        # Field
                        try:
                            valid_field = Feed.verify_data_field_string(field)
                        except ValueError:
                            valid_field = False
                        _field_string = field.rjust(field_str_len)
                        if not valid_field:
                            _field_string = Fore.YELLOW + _field_string + Style.RESET_ALL
                        desc_substring += _field_string + " | "

                        # Last Seen [s ago]
                        _t_diff_string = "{:>20.1f}".format(t_diff)
                        if t_diff > 600:
                            _t_diff_string = Fore.RED + _t_diff_string + Style.RESET_ALL
                        desc_substring += _t_diff_string + " | "

                        # Seen At [ctime]
                        _t_last_string = "{:>20}".format(d_info['t_last'])
                        desc_substring += _t_last_string + " | "

                        # Value
                        _v_last_string = "{:>20}".format(d_info['v_last'])
                        desc_substring += _v_last_string

                        description_string += desc_substring + "\n"

            description_string += "\n"

        return description_string


def main(target, verbose=False):
    checker = DataChecker(target, verbose)
    checker.run()
    print(checker)



from ocs import ocs_agent, site_config
from ocs.base import OpCode
import time
from twisted.internet.defer import inlineCallbacks
from autobahn.twisted.util import sleep as dsleep
from ocs.ocs_feed import Feed
import argparse


class RegisteredAgent:
    """
        Contains data about registered agents.

        Args:
            feed (dict): Encoded :class:`ocs.ocs_feed.Feed`.

        Attributes:
            expired (bool):
                True if agent has not been updated in Registry.agent_timeout seconds.
            time_expired (float, optional):
                ctime at which the agent expired. This will be None if the agent
                is not expired.
            last_updated (float):
                ctime at which the agent was last updated
            op_codes (dict):
                Dictionary of operation codes for each of the agent's
                operations. For details on what the operation codes mean, see
                docs from the ``ocs_agent`` module
    """

    def __init__(self, feed):
        self.expired = False
        self.time_expired = None
        self.last_updated = time.time()
        self.op_codes = {}
        self.agent_class = feed.get('agent_class')
        self.agent_address = feed['agent_address']

    def refresh(self, op_codes=None):
        self.expired = False
        self.time_expired = None
        self.last_updated = time.time()

        if op_codes:
            self.op_codes.update(op_codes)

    def expire(self):
        self.expired = True
        self.time_expired = time.time()
        for k in self.op_codes:
            self.op_codes[k] = OpCode.EXPIRED.value

    def encoded(self):
        return {
            'expired': self.expired,
            'time_expired': self.time_expired,
            'last_updated': self.last_updated,
            'op_codes': self.op_codes,
            'agent_class': self.agent_class,
            'agent_address': self.agent_address,
        }


class Registry:
    """
        The Registry agent is in charge of keeping track of which agents are
        currently running. It has a single process "main" that loops and keeps track
        of when agents expire. This agent subscribes to all heartbeat feeds,
        so no additional function calls are required to register an agent.

        A list of agent statuses is maintained in the "main" process's session.data
        object.

        Args:
            agent (OCSAgent):
                the ocs agent object

        Attributes:
            registered_agents (defaultdict):
                A defaultdict of RegisteredAgent objects, which contain whether
                the agent has expired, the time_expired, and the last_updated
                time.
            agent_timeout (float):
                The time an agent has between heartbeats before being marked
                as expired.
    """

    def __init__(self, agent, args):
        self.log = agent.log
        self.agent = agent
        self.wait_time = args.wait_time

        # Tracking for 'main' Process
        self._run = False

        # Dict containing agent_data for each registered agent
        self.registered_agents = {}
        self.agent_timeout = 5.0  # Removes agent after 5 seconds of no heartbeat.

        self.agent.subscribe_on_start(
            self._register_heartbeat, f'{args.address_root}..feeds.heartbeat',
            options={'match': 'wildcard'}
        )

        agg_params = {
            'frame_length': 60,
        }
        self.agent.register_feed('agent_operations', record=True,
                                 agg_params=agg_params, buffer_time=0)

    def _register_heartbeat(self, _data):
        """
            Function that is called whenever a heartbeat is received from an agent.
            It will update that agent in the Registry's registered_agent dict.
        """
        op_codes, feed = _data
        addr = feed['agent_address']
        if addr not in self.registered_agents:
            self.registered_agents[addr] = RegisteredAgent(feed)

        reg_agent = self.registered_agents[addr]
        publish = op_codes != reg_agent.op_codes
        self.registered_agents[addr].refresh(op_codes=op_codes)
        if publish:
            self._publish_agent_ops(reg_agent)

    def _publish_agent_ops(self, reg_agent):
        """Publish a registered agent's OpCodes.

        Args:
            reg_agent (RegisteredAgent): The registered agent.

        """
        addr = reg_agent.agent_address
        self.log.debug(addr)
        for op_name, op_code in reg_agent.op_codes.items():
            field = f'{addr}_{op_name}'
            field = field.replace('.', '_')
            field = field.replace('-', '_')
            field = Feed.enforce_field_name_rules(field)
            try:
                Feed.verify_data_field_string(field)
            except ValueError as e:
                self.log.warn(f"Improper field name: {field}\n{e}")
                continue
            msg = {'block_name': field,
                   'timestamp': time.time(),
                   'data': {field: op_code}}
            self.agent.publish_to_feed('agent_operations', msg)

    @ocs_agent.param('test_mode', default=False, type=bool)
    @inlineCallbacks
    def main(self, session: ocs_agent.OpSession, params):
        """main(test_mode=False)

        **Process** - Main run process for the Registry agent. This will loop
        and keep track of which agents have expired. It will keep track of
        current active agents in the session.data variable so it can be seen by
        clients.

        Parameters:
            test_mode (bool, optional): Run the main Process loop only once.
                This is meant only for testing. Default is False.

        Notes:
            The session data object for this process will be a dictionary containing
            the encoded RegisteredAgent objects for each agent observed during the
            lifetime of the Registry. For instance, this might look like::

                >>> response.session['data']
                {
                  "observatory.aggregator": {
                    "expired": False,
                    "time_expired": None,
                    "last_updated": 1669925713.4082503,
                    "op_codes": {
                      "record": 3
                    },
                    "agent_class": "AggregatorAgent",
                    "agent_address": "observatory.aggregator"
                  },
                  "observatory.fake-hk-agent-01": {
                    "expired": False,
                    "time_expired": None,
                    "last_updated": 1669925945.7575383,
                    "op_codes": {
                      "acq": 3,
                      "set_heartbeat": 1,
                      "delay_task": 1
                    },
                    "agent_class": "FakeDataAgent",
                    "agent_address": "observatory.fake-hk-agent-01"
                  }
                }

        """
        self._run = True

        last_publish = time.time()
        while self._run:
            yield dsleep(1)

            now = time.time()
            for k, agent in self.registered_agents.items():
                if now - agent.last_updated > self.agent_timeout:
                    agent.expire()

            session.data = {
                k: agent.encoded() for k, agent in self.registered_agents.items()
            }

            if now - last_publish >= self.wait_time:
                last_publish = now
                for agent in self.registered_agents.values():
                    self._publish_agent_ops(agent)

            if params['test_mode']:
                break

        return True, "Stopped registry main process"

    @inlineCallbacks
    def _stop_main(self, session, params):
        """Stop function for the 'main' process."""
        yield
        if self._run:
            session.set_status('stopping')
            self._run = False
            return True, 'requested to stop main process'
        else:
            return False, 'main process not currently running'

    def _register_agent(self, session, agent_data):
        self.log.warn(
            "Warning!!! The register_agent task has been deprecated. Agent '{}' "
            "is using an out of date version of ocs or socs!!"
            .format(agent_data['agent_address'])
        )

        return True, "'register_agent' is deprecated"


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--wait-time', type=float, default=30.,
                        help='Sleep time for main loop')
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='RegistryAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    registry = Registry(agent, args)

    agent.register_process('main', registry.main, registry._stop_main, blocking=False, startup=True)
    agent.register_task('register_agent', registry._register_agent, blocking=False)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()
